ERROR_MESSAGES = {
    "JIRA_ISSUE_KEY_REQUIRED": "Jira issue key is required.",
    "JIRA_MARKDOWN_PATH_REQUIRED": "Markdown path is required.",
    "JIRA_MARKDOWN_NOT_FOUND": "Markdown file not found: {markdown_path}",
    "JIRA_GET_ISSUE_FAILED": "Failed to fetch Jira issue '{issue_key}'.",
    "JIRA_GET_REMOTE_LINKS_FAILED": "Failed to fetch Jira remote links for '{issue_key}'.",
    "JIRA_ADD_COMMENT_FAILED": "Failed to add Jira comment for '{issue_key}'.",
    "JIRA_ADD_ATTACHMENT_FAILED": "Failed to add Jira attachment for '{issue_key}'.",
    "JIRA_MERMAID_RENDER_FAILED": "Failed to render Mermaid diagram for Jira comment. Install `mmdc` (npm @mermaid-js/mermaid-cli) or Docker.",
    "JIRA_COMMENT_BODY_TOO_LONG": "Jira comment body exceeds maximum allowed length ({max_length} characters). Current length: {current_length} characters.",
}


def get_jira_error_message(message_key: str, **kwargs: str) -> str:
    template = ERROR_MESSAGES.get(message_key, "Unknown Jira error.")
    return template.format(**kwargs)


if __name__ == "__main__":
    print(get_jira_error_message("JIRA_ISSUE_KEY_REQUIRED"))



from typing import TypedDict

from core.logging import LocalLogging
from utils.file_manager import FileManager


class ParsedNodeGroupEntry(TypedDict):
    key: str
    entry_start: int
    entry_end: int
    role: str

class TerraformClient:
    def __init__(self) -> None:
        self.logger = LocalLogging.get_logger("hape.terraform_client")
        self.file_manager = FileManager()

if __name__ == "__main__":
    pass


import re

from datetime import datetime


class DatetimeUtils:
    @staticmethod
    def normalize_date_format(date: str, reversed: bool = False) -> str:
        valid_date_regex = re.compile(
            r"\b(0[1-9]|[12]\d|3[01])([./-])(0[1-9]|1[0-2])\2(\d{4})\b"
        )
        date_like_regex = re.compile(r"\b\d{1,2}[./-]\d{1,2}[./-]\d{4}\b")
        if reversed:
            normalized = valid_date_regex.sub(r"\4-\3-\1", date)
        else:
            normalized = valid_date_regex.sub(r"\1-\3-\4", date)
        if not date_like_regex.search(normalized):
            raise ValueError(
                "Invalid date format; expected dd-mm-yyyy, dd.mm.yyyy or dd/mm/yyyy"
            )

        return normalized

    @staticmethod
    def date_to_isoformat(date_input: str) -> str:
        normalized = DatetimeUtils.normalize_date_format(date_input)
        return datetime.strptime(normalized, "%d-%m-%Y").isoformat()

    @staticmethod
    def parse_iso_datetime(value: str) -> datetime:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))



import argparse
import sys
from importlib import metadata

from core.config import Config
from core.errors.handler import ErrorHandler
from cli.commands.config_commands import ConfigCommands
from cli.commands.confluence_commands import ConfluenceCommands
from cli.commands.csv_commands import CsvCommands
from cli.commands.eks_deployment_cost_commands import EksDeploymentCostCommands
from cli.commands.gitlab_commands import GitLabCommands
from cli.commands.jira_commands import JiraCommands
from cli.commands.markdown_commands import MarkdownCommands

class _CommandHelpFormatter(argparse.HelpFormatter):
    def _format_action(self, action) -> str:
        if isinstance(action, argparse._SubParsersAction):
            formatted = super()._format_action(action)
            lines = formatted.splitlines()
            if len(lines) > 1:
                return "\n".join(lines[1:]) + "\n"
        return super()._format_action(action)

class _HapeArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> None:
        self.print_help(sys.stderr)
        self.exit(2, f"\nerror: {message}\n")

class CLI:
    @staticmethod
    def _get_version() -> str:
        try:
            return metadata.version("hape")
        except metadata.PackageNotFoundError:
            return "unknown"

    @staticmethod
    def build_parser() -> argparse.ArgumentParser:
        version = CLI._get_version()
        parser = _HapeArgumentParser(
            prog="hape",
            description="CLI for platform and DevOps automations.",
            usage="%(prog)s [-h] [command] ...",
            formatter_class=_CommandHelpFormatter,
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"hape {version}",
            help="print the installed hape version and exit.",
        )
        parser.add_argument(
            "--config-file-path",
            required=False,
            default=None,
            help="path to config.json (default: ~/.hape/config.json).",
        )
        parser._positionals.title = "commands"
        subparsers = parser.add_subparsers(
            dest="command",
            metavar="command",
            parser_class=_HapeArgumentParser,
        )
        subparsers.required = False

        ConfigCommands.register(subparsers)
        GitLabCommands.register(subparsers)
        JiraCommands.register(subparsers)
        ConfluenceCommands.register(subparsers)
        CsvCommands.register(subparsers)
        EksDeploymentCostCommands.register(subparsers)
        MarkdownCommands.register(subparsers)

        return parser

    @staticmethod
    def run() -> None:
        parser = CLI.build_parser()
        if len(sys.argv) == 1:
            parser.print_help()
            return
        args = parser.parse_args()
        Config.set_config_path(args.config_file_path)
        if not args.command:
            parser.print_help()
            return
        try:
            args.func(args)
        except Exception as exc:
            exit_code = ErrorHandler.handle(exc)
            raise SystemExit(exit_code)

def main() -> None:
    CLI.run()

if __name__ == "__main__":
    main()



import base64

import requests

from core.config import Config
from core.logging import LocalLogging
from typing import Optional, List
from utils.urls_utils import UrlsUtils


class ConfluenceClient:
    def __init__(self):
        self.logger = LocalLogging.get_logger("hape.confluence_client")
        self.confluence_email = Config.get_atlassian_email()
        self.confluence_api_token = Config.get_atlassian_api_key()
        raw_domain = Config.get_atlassian_domain()
        base_url = UrlsUtils.normalize_atlassian_base_url(raw_domain)
        self.confluence_base_url = self._ensure_wiki_base(base_url)
        self.changelog_parent_page_id = Config.get_changelog_parent_page_id()
        self.changelog_entry_page_template_id = Config.get_changelog_entry_page_template_id()
        self.test_parent_page_id = Config.get_test_parent_page_id()

        if not self.confluence_email:
            raise ValueError("ATLASSIAN_EMAIL is required to access Confluence.")
        if not self.confluence_api_token:
            raise ValueError("ATLASSIAN_API_KEY is required to access Confluence.")

    def _get_auth_headers(self):
        credentials = f"{self.confluence_email}:{self.confluence_api_token}"
        encoded = base64.b64encode(credentials.encode("utf-8")).decode("ascii")
        return {
            "Authorization": f"Basic {encoded}",
            "Accept": "application/json",
        }

    def _ensure_wiki_base(self, base_url):
        normalized = base_url.rstrip("/")
        if normalized.endswith("/wiki"):
            return normalized
        return f"{normalized}/wiki"

    def get_page(self, page_id: str, expand: Optional[str] = None) -> dict:
        self.logger.debug(f"get_page(page_id: {page_id}, expand: {expand})")
        page_url = f"{self.confluence_base_url}/rest/api/content/{page_id}"
        params: dict[str, str] = {}
        if expand:
            params["expand"] = expand
        response = requests.get( page_url, params=params, headers=self._get_auth_headers())
        response.raise_for_status()
        return response.json()

    def create_page(self, parent_page_id: str, page_title: str, page_body: str, space_key: str, labels: Optional[List[str]] = None, set_full_width: bool = True) -> dict:
        self.logger.debug(f"create_page(parent_page_id: {parent_page_id}, page_title: {page_title}, space_key: {space_key})")
        page_url = f"{self.confluence_base_url}/rest/api/content"
        payload = {
            "type": "page",
            "title": page_title,
            "space": {"key": space_key},
            "ancestors": [{"id": str(parent_page_id)}],
            "body": {
                "storage": {
                    "value": page_body,
                    "representation": "storage",
                }
            },
        }
        if labels is not None and isinstance(labels, list):
            label_list = []
            for label in labels:
                label_list.append({"prefix": "global", "name": label})
            payload["metadata"] = {"labels": label_list}
        headers = {
            **self._get_auth_headers(),
            "Content-Type": "application/json",
        }
        response = requests.post(page_url, headers=headers, json=payload)
        response.raise_for_status()
        new_page = response.json()
        if set_full_width:
            self.set_page_full_width(new_page["id"])
        return new_page

    def set_page_full_width(self, page_id):
        self.logger.debug(f"set_page_full_width(page_id: {page_id})")
        properties_url = f"{self.confluence_base_url}/api/v2/pages/{page_id}/properties"
        headers = {
            **self._get_auth_headers(),
            "Content-Type": "application/json",
        }
        for key in ("content-appearance-draft", "content-appearance-published"):
            response = requests.post(properties_url, headers=headers, json={"key": key, "value": "full-width"})
            response.raise_for_status()



import base64

import requests

from core.config import Config
from core.logging import LocalLogging
from utils.urls_utils import UrlsUtils


class JiraClient:
    def __init__(self):
        self.logger = LocalLogging.get_logger("hape.jira_client")
        self.jira_email = Config.get_atlassian_email()
        self.jira_api_token = Config.get_atlassian_api_key()
        raw_domain = Config.get_atlassian_domain()
        self.jira_base_url = UrlsUtils.normalize_atlassian_base_url(raw_domain)

        if not self.jira_email:
            raise ValueError("ATLASSIAN_EMAIL is required to access Jira.")
        if not self.jira_api_token:
            raise ValueError("ATLASSIAN_API_KEY is required to access Jira.")

    def _get_auth_headers(self):
        credentials = f"{self.jira_email}:{self.jira_api_token}"
        encoded = base64.b64encode(credentials.encode("utf-8")).decode("ascii")
        return {
            "Authorization": f"Basic {encoded}",
            "Accept": "application/json",
        }

    def get_issue(self, issue_key):
        self.logger.debug(f"get_issue(issue_key: {issue_key})")
        issue_url = f"{self.jira_base_url}/rest/api/3/issue/{issue_key}"
        response = requests.get(issue_url, headers=self._get_auth_headers())
        if response.status_code != 200:
            raise RuntimeError(
                "Failed to fetch Jira issue "
                f"'{issue_key}' (status={response.status_code}): {response.text}"
            )
        return response.json()

    def get_issue_remote_links(self, issue_key):
        self.logger.debug(f"get_issue_remote_links(issue_key: {issue_key})")
        links_url = f"{self.jira_base_url}/rest/api/3/issue/{issue_key}/remotelink"
        response = requests.get(links_url, headers=self._get_auth_headers())
        if response.status_code != 200:
            raise RuntimeError(f"Failed to fetch Jira issue remote links '{issue_key}' (status={response.status_code}): {response.text}")
        return response.json()

    def add_comment(self, issue_key: str, comment_body: str) -> dict:
        self.logger.debug(f"add_comment(issue_key: {issue_key})")
        comment_url = f"{self.jira_base_url}/rest/api/2/issue/{issue_key}/comment"
        headers = {
            **self._get_auth_headers(),
            "Content-Type": "application/json",
        }
        response = requests.post(
            comment_url,
            headers=headers,
            json={"body": comment_body},
        )
        response.raise_for_status()
        return response.json()

    def add_attachment(self, issue_key: str, filename: str, content_bytes: bytes, content_type: str = "application/octet-stream") -> list:
        self.logger.debug(f"add_attachment(issue_key: {issue_key}, filename: {filename})")
        attachment_url = f"{self.jira_base_url}/rest/api/2/issue/{issue_key}/attachments"
        headers = {
            **self._get_auth_headers(),
            "X-Atlassian-Token": "no-check",
        }
        files = {
            "file": (filename, content_bytes, content_type),
        }
        response = requests.post(attachment_url, headers=headers, files=files)
        response.raise_for_status()
        return response.json()

    def get_issue_url(self, issue_key: str) -> str:
        return f"{self.jira_base_url}/browse/{issue_key}"



import argparse
from typing import Any

from services.csv_service import CsvService


class CsvCommands:
    @staticmethod
    def register(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            "csv",
            help="csv conversion operations.",
        )
        parser.set_defaults(func=CsvCommands.run_help, parser=parser)
        csv_subparsers = parser.add_subparsers(
            dest="csv_command",
            metavar="command",
        )
        csv_subparsers.required = False

        from_json_parser = csv_subparsers.add_parser(
            "from-json",
            help="convert JSON input to CSV file.",
        )
        from_json_parser.add_argument(
            "--json-file",
            required=False,
            default=None,
            help="path to JSON file input.",
        )
        from_json_parser.add_argument(
            "--json-content",
            required=False,
            default=None,
            help="raw JSON array input string.",
        )
        from_json_parser.add_argument(
            "--output-file",
            required=True,
            default=None,
            help="path to CSV output file.",
        )
        from_json_parser.add_argument(
            "--delimiter",
            required=False,
            default=",",
            help="CSV delimiter (default: ,).",
        )
        from_json_parser.set_defaults(func=CsvCommands.run_from_json)

        to_json_parser = csv_subparsers.add_parser(
            "to-json",
            help="convert CSV file to JSON output file.",
        )
        to_json_parser.add_argument(
            "--csv-file",
            required=True,
            default=None,
            help="path to CSV input file.",
        )
        to_json_parser.add_argument(
            "--output-file",
            required=True,
            default=None,
            help="path to JSON output file.",
        )
        to_json_parser.add_argument(
            "--delimiter",
            required=False,
            default=",",
            help="CSV delimiter (default: ,).",
        )
        to_json_parser.set_defaults(func=CsvCommands.run_to_json)

    @staticmethod
    def run_from_json(args: Any) -> None:
        csv_service = CsvService()
        output_file = csv_service.from_json(
            output_file=args.output_file,
            json_file=args.json_file,
            json_content=args.json_content,
            delimiter=args.delimiter,
        )
        print(output_file)

    @staticmethod
    def run_to_json(args: Any) -> None:
        csv_service = CsvService()
        output_file = csv_service.to_json(
            csv_file=args.csv_file,
            output_file=args.output_file,
            delimiter=args.delimiter,
        )
        print(output_file)

    @staticmethod
    def run_help(args: Any) -> None:
        args.parser.print_help()



import argparse
import json
from typing import Any

from core.logging import LocalLogging
from services.jira_service import JiraService


class JiraCommands:
    @staticmethod
    def register(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            "jira",
            help="fetch Jira issue data, remote links, or add comments.",
        )
        parser.set_defaults(func=JiraCommands.run_help, parser=parser)
        jira_subparsers = parser.add_subparsers(
            dest="jira_command",
            metavar="command",
        )
        jira_subparsers.required = False

        convert_parser = jira_subparsers.add_parser(
            "md-to-comment",
            help="add a Jira comment from a markdown file.",
        )
        convert_parser.add_argument(
            "--issue-key",
            required=True,
            default=None,
            help="issue key.",
        )
        convert_parser.add_argument(
            "--md-path",
            required=True,
            default=None,
            help="path to markdown file to post as a comment.",
        )
        convert_parser.set_defaults(func=JiraCommands.run_convert_md_to_comment)

    @staticmethod
    def run_convert_md_to_comment(args: Any) -> None:
        LocalLogging.bootstrap()
        jira_service = JiraService()
        data = jira_service.add_comment_from_markdown(
            issue_key=args.issue_key,
            markdown_path=args.md_path,
        )
        issue_url = jira_service.get_issue_url(args.issue_key)
        print(issue_url)

    @staticmethod
    def run_help(args: Any) -> None:
        args.parser.print_help()



import argparse
from typing import Any

from core.logging import LocalLogging
from services.config_service import ConfigService


class ConfigCommands:
    @staticmethod
    def register(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            "config",
            help="config file operations.",
        )
        parser.set_defaults(func=ConfigCommands.run_help, parser=parser)
        config_subparsers = parser.add_subparsers(
            dest="config_command",
            metavar="command",
        )
        config_subparsers.required = False

        init_parser = config_subparsers.add_parser(
            "init-config-file",
            help="generate config.json from .env.",
        )
        init_parser.add_argument(
            "--dot-env-file",
            required=False,
            default=None,
            help="path to .env file (optional). If omitted, use environment variables.",
        )
        init_parser.set_defaults(func=ConfigCommands.run_init_config_file)

    @staticmethod
    def run_init_config_file(args: Any) -> None:
        LocalLogging.bootstrap()
        config_service = ConfigService()
        config_path = config_service.init_config_file(
            config_path=args.config_file_path,
            dot_env_file=args.dot_env_file,
        )
        print(config_path)

    @staticmethod
    def run_help(args: Any) -> None:
        args.parser.print_help()



import os
from typing import List, Optional

from clients.confluence_client import ConfluenceClient
from core.config import Config
from core.errors.exceptions import HapeExternalError, HapeOperationError, HapeValidationError
from core.errors.messages.confluence_error_messages import get_confluence_error_message
from core.logging import LocalLogging
from utils.formatters.confluence_doc_formatter import ConfluenceDocFormatter
from utils.file_manager import FileManager


class ConfluenceService:
    def __init__(self, confluence_client: Optional[ConfluenceClient] = None) -> None:
        self.confluence_client = confluence_client or ConfluenceClient()
        self.logger = LocalLogging.get_logger("hape.confluence_service")
        self.file_manager = FileManager()
        
    def _markdown_to_storage(self, markdown_text: str) -> str:
        return ConfluenceDocFormatter.format_markdown_to_storage(markdown_text)

    def _derive_page_title(self, markdown_text: str, readme_path: str) -> str:
        for line in markdown_text.splitlines():
            if line.startswith("# "):
                title = line.replace("# ", "", 1).strip()
                if title:
                    return title
                break
        fallback = os.path.splitext(os.path.basename(readme_path))[0].strip()
        return fallback or "README"    

    def get_page(self, page_id: str, expand: str = "body.storage") -> dict:
        self.logger.debug(f"get_page(page_id: {page_id}, expand: {expand})")
        if not page_id or not page_id.strip():
            raise HapeValidationError(
                code="CONFLUENCE_PAGE_ID_REQUIRED",
                message=get_confluence_error_message("CONFLUENCE_PAGE_ID_REQUIRED"),
            )
        try:
            return self.confluence_client.get_page(page_id, expand=expand)
        except Exception as exc:
            raise HapeExternalError(
                code="CONFLUENCE_GET_PAGE_FAILED",
                message=get_confluence_error_message(
                    "CONFLUENCE_GET_PAGE_FAILED",
                    page_id=page_id,
                ),
            ) from exc

    def create_page(self, parent_page_id: str, page_title: str, page_body: str, space_key: str, labels: Optional[List[str]] = None) -> dict:
        self.logger.debug(
            "create_page(parent_page_id: {parent_page_id}, page_title: {page_title}, space_key: {space_key})")
        if not parent_page_id or not str(parent_page_id).strip():
            raise HapeValidationError(
                code="CONFLUENCE_PARENT_PAGE_ID_REQUIRED",
                message=get_confluence_error_message("CONFLUENCE_PARENT_PAGE_ID_REQUIRED"),
            )
        if not page_title or not page_title.strip():
            raise HapeValidationError(
                code="CONFLUENCE_PAGE_TITLE_REQUIRED",
                message=get_confluence_error_message("CONFLUENCE_PAGE_TITLE_REQUIRED"),
            )
        if not page_body or not page_body.strip():
            raise HapeValidationError(
                code="CONFLUENCE_PAGE_BODY_REQUIRED",
                message=get_confluence_error_message("CONFLUENCE_PAGE_BODY_REQUIRED"),
            )
        try:
            return self.confluence_client.create_page(
                parent_page_id=parent_page_id,
                page_title=page_title,
                page_body=page_body,
                space_key=space_key,
                labels=labels,
            )
        except Exception as exc:
            raise HapeExternalError(
                code="CONFLUENCE_CREATE_PAGE_FAILED",
                message=get_confluence_error_message(
                    "CONFLUENCE_CREATE_PAGE_FAILED",
                    page_title=page_title,
                ),
            ) from exc

    def get_page_link(self, page_id: str, space_key: str) -> str:
        base_url = self.confluence_client.confluence_base_url
        return f"{base_url}/spaces/{space_key}/pages/{page_id}"

    def create_page_from_markdown(self, readme_path: str, space_key: str, parent_page_id: Optional[str] = None, page_title: Optional[str] = None, labels: Optional[List[str]] = None) -> dict:
        self.logger.debug(f"create_page_from_markdown(parent_page_id: {parent_page_id}, readme_path: {readme_path}, space_key: {space_key})")
        if not readme_path or not readme_path.strip():
            raise HapeValidationError(
                code="CONFLUENCE_README_PATH_REQUIRED",
                message=get_confluence_error_message("CONFLUENCE_README_PATH_REQUIRED"),
            )
        if not self.file_manager.file_exists(readme_path):
            raise HapeValidationError(
                code="CONFLUENCE_README_NOT_FOUND",
                message=get_confluence_error_message(
                    "CONFLUENCE_README_NOT_FOUND",
                    readme_path=readme_path,
                ),
            )
        markdown_text = self.file_manager.read_file(readme_path)
        page_title = page_title or self._derive_page_title(markdown_text, readme_path)
        page_title = "[AUTOGENERATED] " + page_title
        try:
            page_body = self._markdown_to_storage(markdown_text)
        except Exception as exc:
            raise HapeOperationError(
                code="CONFLUENCE_MARKDOWN_CONVERSION_FAILED",
                message=get_confluence_error_message("CONFLUENCE_MARKDOWN_CONVERSION_FAILED"),
            ) from exc
        parent_page_id = parent_page_id or Config.get_test_parent_page_id()
        return self.create_page(
            parent_page_id=parent_page_id,
            page_title=page_title,
            page_body=page_body,
            space_key=space_key,
            labels=labels,
        )



import json
import time
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from threading import Lock
from typing import Any

from core.config import Config
from core.logging import LocalLogging
from services.eks_deployment_cost_service import EksDeploymentCostService

METRICS_CATALOG = [
    {
        "name": "hape_eks_deployment_cost_exporter_up",
        "type": "gauge",
        "description": "1 when exporter refresh succeeds.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_exporter_last_refresh_timestamp_seconds",
        "type": "gauge",
        "description": "Last successful or failed refresh timestamp.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_workload_count",
        "type": "gauge",
        "description": "Count of discovered workloads.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_total_cpu_request_cores",
        "type": "gauge",
        "description": "Total CPU requests across workloads.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_total_memory_request_gib",
        "type": "gauge",
        "description": "Total memory requests across workloads.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_total_hourly_cpu_usd",
        "type": "gauge",
        "description": "Estimated hourly CPU cost.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_total_hourly_memory_usd",
        "type": "gauge",
        "description": "Estimated hourly memory cost.",
        "labels": [],
    },
    {
        "name": "hape_eks_deployment_cost_total_usd",
        "type": "gauge",
        "description": "Estimated total cost by period.",
        "labels": ["period=hourly|daily|monthly|yearly"],
    },
    {
        "name": "hape_eks_deployment_cost_workload_max_hourly_usd",
        "type": "gauge",
        "description": "Estimated workload hourly cost (max CPU or memory).",
        "labels": ["resource_type", "namespace", "name", "instance_type"],
    },
]
METRICS_CATALOG_BY_NAME = {metric["name"]: metric for metric in METRICS_CATALOG}


def _escape_label(value: str) -> str:
    return value.replace("\\", "\\\\").replace("\n", "\\n").replace('"', '\\"')


def _metric_line(name: str, value: float | int, labels: dict[str, str] | None = None) -> str:
    if not labels:
        return f"{name} {value}"
    label_parts = [f'{key}="{_escape_label(label_value)}"' for key, label_value in labels.items()]
    return f'{name}{{{",".join(label_parts)}}} {value}'


def _metric_metadata_lines(metric_name: str) -> list[str]:
    metric = METRICS_CATALOG_BY_NAME.get(metric_name)
    if metric is None:
        raise ValueError(f"Metric metadata is missing for '{metric_name}'.")
    return [f"# HELP {metric_name} {metric['description']}", f"# TYPE {metric_name} {metric['type']}"]


class EksDeploymentCostMetricsProvider:
    DEFAULT_REFRESH_SECONDS = 300

    def __init__(
        self,
        refresh_seconds: int = DEFAULT_REFRESH_SECONDS,
        kube_context: str | None = None,
        aws_profile: str | None = None,
        ignored_namespaces_csv: str | None = None,
    ) -> None:
        self.refresh_seconds = refresh_seconds
        self.kube_context = kube_context
        self.aws_profile = aws_profile
        self.ignored_namespaces_csv = ignored_namespaces_csv
        self.service = EksDeploymentCostService()
        self.logger = LocalLogging.get_logger("hape.eks_deployment_cost_metrics_provider")
        self._lock = Lock()
        self._last_refresh_epoch_seconds = 0.0
        self._last_metrics_payload = ""
        self._last_error = ""

    def _refresh(self) -> None:
        self.logger.debug("Refreshing exporter metrics.")
        try:
            report_file_paths = self.service.generate_report(
                kube_context=self.kube_context,
                aws_profile=self.aws_profile,
                ignored_namespaces_csv=self.ignored_namespaces_csv,
                use_incluster_config=self.kube_context is None,
            )
            summary_path = Path(report_file_paths["summary_json_path"])
            summary_report = json.loads(summary_path.read_text(encoding="utf-8"))
            self._last_metrics_payload = self._build_payload(summary_report=summary_report, exporter_up=1)
            self._last_error = ""
            self.logger.debug("Exporter metrics refreshed successfully.")
        except Exception as exc:
            self._last_error = str(exc)
            self.logger.exception("Failed to refresh EKS deployment cost metrics.")
            self._last_metrics_payload = self._build_payload(summary_report=None, exporter_up=0)
        self._last_refresh_epoch_seconds = time.time()

    def _build_payload(self, summary_report: dict[str, Any] | None, exporter_up: int) -> str:
        lines = [
            *_metric_metadata_lines("hape_eks_deployment_cost_exporter_up"),
            _metric_line("hape_eks_deployment_cost_exporter_up", exporter_up),
            *_metric_metadata_lines("hape_eks_deployment_cost_exporter_last_refresh_timestamp_seconds"),
            _metric_line("hape_eks_deployment_cost_exporter_last_refresh_timestamp_seconds", int(time.time())),
        ]
        if not summary_report:
            lines.extend(
                [
                    *_metric_metadata_lines("hape_eks_deployment_cost_workload_count"),
                    _metric_line("hape_eks_deployment_cost_workload_count", 0),
                ]
            )
            return "\n".join(lines) + "\n"

        summary = summary_report.get("summary", {})
        top_workloads = summary_report.get("top_costing_workloads", [])

        lines.extend(
            [
                *_metric_metadata_lines("hape_eks_deployment_cost_workload_count"),
                _metric_line("hape_eks_deployment_cost_workload_count", int(summary.get("workload_count", 0))),
                *_metric_metadata_lines("hape_eks_deployment_cost_total_cpu_request_cores"),
                _metric_line("hape_eks_deployment_cost_total_cpu_request_cores", float(summary.get("total_cpu_request_cores", 0.0))),
                *_metric_metadata_lines("hape_eks_deployment_cost_total_memory_request_gib"),
                _metric_line("hape_eks_deployment_cost_total_memory_request_gib", float(summary.get("total_memory_request_gib", 0.0))),
                *_metric_metadata_lines("hape_eks_deployment_cost_total_hourly_cpu_usd"),
                _metric_line("hape_eks_deployment_cost_total_hourly_cpu_usd", float(summary.get("hourly_cpu_cost_usd", 0.0))),
                *_metric_metadata_lines("hape_eks_deployment_cost_total_hourly_memory_usd"),
                _metric_line("hape_eks_deployment_cost_total_hourly_memory_usd", float(summary.get("hourly_memory_cost_usd", 0.0))),
                *_metric_metadata_lines("hape_eks_deployment_cost_total_usd"),
                _metric_line("hape_eks_deployment_cost_total_usd", float(summary.get("max_hourly_cost_usd", 0.0)), labels={"period": "hourly"}),
                _metric_line("hape_eks_deployment_cost_total_usd", float(summary.get("daily_cost_usd", 0.0)), labels={"period": "daily"}),
                _metric_line("hape_eks_deployment_cost_total_usd", float(summary.get("monthly_cost_usd", 0.0)), labels={"period": "monthly"}),
                _metric_line("hape_eks_deployment_cost_total_usd", float(summary.get("yearly_cost_usd", 0.0)), labels={"period": "yearly"}),
                *_metric_metadata_lines("hape_eks_deployment_cost_workload_max_hourly_usd"),
            ]
        )
        for workload in top_workloads:
            labels = {
                "resource_type": str(workload.get("resource_type", "")),
                "namespace": str(workload.get("namespace", "")),
                "name": str(workload.get("name", "")),
                "instance_type": str(workload.get("instance_type", "")),
            }
            lines.append(
                _metric_line(
                    "hape_eks_deployment_cost_workload_max_hourly_usd",
                    float(workload.get("max_hourly_cost_usd", 0.0)),
                    labels=labels,
                )
            )
        return "\n".join(lines) + "\n"

    def get_metrics_payload(self) -> str:
        now = time.time()
        is_stale = (now - self._last_refresh_epoch_seconds) >= self.refresh_seconds
        if is_stale or not self._last_metrics_payload:
            with self._lock:
                now_locked = time.time()
                is_stale_locked = (now_locked - self._last_refresh_epoch_seconds) >= self.refresh_seconds
                if is_stale_locked or not self._last_metrics_payload:
                    self._refresh()
        return self._last_metrics_payload

    def get_last_error(self) -> str:
        return self._last_error

    def get_metrics_catalog_json(self) -> str:
        return json.dumps({"metrics": METRICS_CATALOG}, indent=2) + "\n"


def make_handler(provider: EksDeploymentCostMetricsProvider) -> type[BaseHTTPRequestHandler]:
    class EksDeploymentCostExporterHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            if self.path in ("/", "/catalog", "/metrics-catalog"):
                payload = provider.get_metrics_catalog_json().encode("utf-8")
                self.send_response(200)
                self.send_header("Content-Type", "application/json; charset=utf-8")
                self.send_header("Content-Length", str(len(payload)))
                self.end_headers()
                self.wfile.write(payload)
                return
            if self.path == "/metrics":
                payload = provider.get_metrics_payload().encode("utf-8")
                self.send_response(200)
                self.send_header("Content-Type", "text/plain; version=0.0.4; charset=utf-8")
                self.send_header("Content-Length", str(len(payload)))
                self.end_headers()
                self.wfile.write(payload)
                return
            if self.path == "/healthz":
                error = provider.get_last_error()
                if error:
                    body = "degraded\n".encode("utf-8")
                    self.send_response(503)
                else:
                    body = "ok\n".encode("utf-8")
                    self.send_response(200)
                self.send_header("Content-Type", "text/plain; charset=utf-8")
                self.send_header("Content-Length", str(len(body)))
                self.end_headers()
                self.wfile.write(body)
                return
            self.send_response(404)
            self.end_headers()

        def log_message(self, format_text: str, *args: Any) -> None:
            provider.logger.debug(format_text, *args)

    return EksDeploymentCostExporterHandler


def main() -> None:
    LocalLogging.bootstrap()
    host = Config.get_exporter_host()
    port = Config.get_exporter_port()
    refresh_seconds = Config.get_exporter_refresh_seconds()
    kube_context = Config.get_edc_kube_context() or None
    aws_profile = Config.get_edc_aws_profile() or None
    ignored_namespaces_csv = Config.get_edc_ignored_namespaces_csv()
    provider = EksDeploymentCostMetricsProvider(
        refresh_seconds=refresh_seconds,
        kube_context=kube_context,
        aws_profile=aws_profile,
        ignored_namespaces_csv=ignored_namespaces_csv,
    )
    handler = make_handler(provider=provider)
    server = ThreadingHTTPServer((host, port), handler)
    logger = LocalLogging.get_logger("hape.eks_deployment_cost_exporter")
    logger.info("Starting EKS deployment cost exporter on %s:%s", host, port)
    server.serve_forever()


if __name__ == "__main__":
    main()



import csv
import json
from pathlib import Path
from typing import Any


class CsvManager:

    @staticmethod
    def _normalize_row(fieldnames: list[str], row: dict[str, Any]) -> dict[str, Any]:
        normalized: dict[str, Any] = {}
        for field in fieldnames:
            normalized[field] = row.get(field, "")
        return normalized

    def read_csv(self, csv_path: str, delimiter: str = ",") -> list[dict[str, str]]:
        with open(csv_path, "r", encoding="utf-8", newline="") as csv_file:
            reader = csv.DictReader(csv_file, delimiter=delimiter)
            rows: list[dict[str, str]] = []
            for row in reader:
                clean_row = {str(key).strip(): (value or "").strip() for key, value in row.items()}
                rows.append(clean_row)
        return rows

    def write_csv(self, csv_path: str, rows: list[dict[str, Any]], delimiter: str = ",") -> None:
        if not rows:
            raise ValueError("CSV rows are empty.")
        fieldnames = list(rows[0].keys())
        with open(csv_path, "w", encoding="utf-8", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames, delimiter=delimiter)
            writer.writeheader()
            for row in rows:
                writer.writerow(self._normalize_row(fieldnames, row))

    def parse_json_rows(self, json_text: str) -> list[dict[str, Any]]:
        data = json.loads(json_text)
        if not isinstance(data, list):
            raise ValueError("JSON must be an array of objects.")
        if not data:
            raise ValueError("JSON array is empty.")
        normalized_rows: list[dict[str, Any]] = []
        for row in data:
            if not isinstance(row, dict):
                raise ValueError("Each JSON array item must be an object.")
            normalized_rows.append(row)
        return normalized_rows

    def read_json_rows(self, json_path: str) -> list[dict[str, Any]]:
        content = Path(json_path).read_text(encoding="utf-8")
        return self.parse_json_rows(content)

    def write_json(self, output_path: str, rows: list[dict[str, str]]) -> None:
        Path(output_path).write_text(json.dumps(rows, indent=4), encoding="utf-8")


def _demo_read_write_csv() -> None:
    csv_manager = CsvManager()
    sample_rows = [{"name": "alpha", "value": "1"}, {"name": "beta", "value": "2"}]
    csv_manager.write_csv("/tmp/hape_csv_manager_demo.csv", sample_rows)
    print(csv_manager.read_csv("/tmp/hape_csv_manager_demo.csv"))


def _demo_json_rows() -> None:
    csv_manager = CsvManager()
    json_rows = csv_manager.parse_json_rows('[{"name":"alpha","value":1},{"name":"beta","value":2}]')
    print(json_rows)


if __name__ == "__main__":
    _demo_read_write_csv()
    _demo_json_rows()



import kubernetes.config
import kubernetes.client
import re
import time
from collections import Counter
from typing import Any, Dict, List

from core.logging import LocalLogging
from utils.file_manager import FileManager


class KubernetesClient:
    def __init__(self, context: str | None = None, config_file: str | None = None, use_incluster_config: bool = False) -> None:
        resolved_context = context.strip() if context else ""
        if use_incluster_config:
            kubernetes.config.load_incluster_config()
            resolved_context = "in-cluster"
        elif resolved_context:
            kubernetes.config.load_kube_config(context=resolved_context, config_file=config_file)
        else:
            try:
                kubernetes.config.load_incluster_config()
                resolved_context = "in-cluster"
            except Exception as exc:
                raise ValueError("Error: Kubernetes context is not defined and in-cluster config is unavailable.") from exc

        # Disable TLS verification for this client instance
        configuration = kubernetes.client.Configuration.get_default_copy()
        configuration.verify_ssl = False
        kubernetes.client.Configuration.set_default(configuration)

        self.context = resolved_context
        self.logger = LocalLogging.get_logger("hape.kubernetes_client")
        self.file_manager = FileManager()
        self.core_v1 = kubernetes.client.CoreV1Api()
        self.apps_v1 = kubernetes.client.AppsV1Api()
        self.policy_v1 = kubernetes.client.PolicyV1Api()
        self.autoscaling_v2 = kubernetes.client.AutoscalingV2Api()
        self.apiextensions_v1 = kubernetes.client.ApiextensionsV1Api()
        self.custom_objects = kubernetes.client.CustomObjectsApi()

    def _sanitize_filename(self, name):
        self.logger.debug(f"_sanitize_filename(name: {name})")
        return name.replace("/", "_")

    def _is_node_ready(self, node_name: str) -> bool:
        self.logger.debug(f"_is_node_ready(node_name: {node_name})")
        node = self.core_v1.read_node(node_name)
        conditions = node.status.conditions or []
        for condition in conditions:
            if condition.type == "Ready":
                return condition.status == "True"
        return False

    def _is_pod_ready(self, pod) -> bool:
        self.logger.debug(f"_is_pod_ready(pod: {pod.metadata.name})")
        if pod.status.phase != "Running":
            return False
        conditions = pod.status.conditions or []
        for condition in conditions:
            if condition.type == "Ready":
                return condition.status == "True"
        return False

    def _create_namespaced_pod_eviction(self, namespace: str, pod_name: str, eviction: kubernetes.client.V1Eviction) -> None:
        self.logger.debug(f"_create_namespaced_pod_eviction(namespace: {namespace}, pod_name: {pod_name})")
        if hasattr(self.core_v1, "create_namespaced_pod_eviction"):
            self.core_v1.create_namespaced_pod_eviction(name=pod_name, namespace=namespace, body=eviction)
            return

        if hasattr(self.policy_v1, "create_namespaced_pod_eviction"):
            self.policy_v1.create_namespaced_pod_eviction(name=pod_name, namespace=namespace, body=eviction)
            return

        raise AttributeError("Neither CoreV1Api nor PolicyV1Api exposes create_namespaced_pod_eviction() in installed kubernetes client.")

    @staticmethod
    def _parse_cpu_cores(cpu_text: str | None) -> float:
        if not cpu_text:
            return 0.0
        value = cpu_text.strip()
        if not value:
            return 0.0
        if value.endswith("m"):
            return float(value[:-1]) / 1000.0
        if value.endswith("u"):
            return float(value[:-1]) / 1_000_000.0
        if value.endswith("n"):
            return float(value[:-1]) / 1_000_000_000.0
        return float(value)

    @staticmethod
    def _parse_memory_gib(memory_text: str | None) -> float:
        if not memory_text:
            return 0.0
        value = memory_text.strip()
        if not value:
            return 0.0
        unit_factors = {
            "Ki": 1.0 / (1024.0 ** 2),
            "Mi": 1.0 / 1024.0,
            "Gi": 1.0,
            "Ti": 1024.0,
            "Pi": 1024.0 ** 2,
            "Ei": 1024.0 ** 3,
            "K": 1000.0 / (1024.0 ** 3),
            "M": (1000.0 ** 2) / (1024.0 ** 3),
            "G": (1000.0 ** 3) / (1024.0 ** 3),
            "T": (1000.0 ** 4) / (1024.0 ** 3),
            "P": (1000.0 ** 5) / (1024.0 ** 3),
            "E": (1000.0 ** 6) / (1024.0 ** 3),
        }
        for unit, factor in unit_factors.items():
            if value.endswith(unit):
                return float(value[:-len(unit)]) * factor
        return float(value) / (1024.0 ** 3)

    def _sum_container_requests(self, containers: list) -> tuple[float, float]:
        cpu_total = 0.0
        memory_total = 0.0
        for container in containers or []:
            resources = container.resources
            requests = resources.requests if resources else None
            if not requests:
                continue
            cpu_total += self._parse_cpu_cores(requests.get("cpu"))
            memory_total += self._parse_memory_gib(requests.get("memory"))
        return cpu_total, memory_total

    @staticmethod
    def _build_label_selector(match_labels: dict[str, str]) -> str:
        if not match_labels:
            return ""
        selector_parts = [f"{key}={value}" for key, value in sorted(match_labels.items())]
        return ",".join(selector_parts)

    @staticmethod
    def _get_node_instance_type_label(node) -> str | None:
        labels = node.metadata.labels if node and node.metadata and node.metadata.labels else {}
        return labels.get("node.kubernetes.io/instance-type") or labels.get("beta.kubernetes.io/instance-type")

    def _collect_deployment_request_details(self, namespace: str) -> list[dict[str, Any]]:
        deployments = self.apps_v1.list_namespaced_deployment(namespace=namespace)
        items: list[dict[str, Any]] = []
        for deployment in deployments.items:
            replicas = deployment.spec.replicas
            containers = deployment.spec.template.spec.containers or []
            cpu_request_cores_per_pod, memory_request_gib_per_pod = self._sum_container_requests(containers=containers)
            items.append(
                {
                    "resource_type": "Deployment",
                    "namespace": namespace,
                    "name": deployment.metadata.name,
                    "replicas": replicas,
                    "selector_match_labels": deployment.spec.selector.match_labels or {},
                    "cpu_request_cores_per_pod": cpu_request_cores_per_pod,
                    "memory_request_gib_per_pod": memory_request_gib_per_pod,
                }
            )
        return items

    def _collect_statefulset_request_details(self, namespace: str) -> list[dict[str, Any]]:
        statefulsets = self.apps_v1.list_namespaced_stateful_set(namespace=namespace)
        items: list[dict[str, Any]] = []
        for statefulset in statefulsets.items:
            replicas = statefulset.spec.replicas
            containers = statefulset.spec.template.spec.containers or []
            cpu_request_cores_per_pod, memory_request_gib_per_pod = self._sum_container_requests(containers=containers)
            items.append(
                {
                    "resource_type": "StatefulSet",
                    "namespace": namespace,
                    "name": statefulset.metadata.name,
                    "replicas": replicas,
                    "selector_match_labels": statefulset.spec.selector.match_labels or {},
                    "cpu_request_cores_per_pod": cpu_request_cores_per_pod,
                    "memory_request_gib_per_pod": memory_request_gib_per_pod,
                }
            )
        return items

    @staticmethod
    def list_contexts():
        LocalLogging.get_logger("hape.kubernetes_client").debug("list_contexts()")
        contexts, _ = kubernetes.config.list_kube_config_contexts()
        return [context["name"] for context in contexts or []]

    def list_namespaces(self) -> list[str]:
        self.logger.debug("list_namespaces()")
        response = self.core_v1.list_namespace()
        return [item.metadata.name for item in response.items if item.metadata and item.metadata.name]

    def list_replica_workload_request_details(self, resource_types: list[str], namespaces: list[str] | None = None) -> list[dict[str, Any]]:
        self.logger.debug(f"list_replica_workload_request_details(resource_types: {resource_types}, namespaces: {namespaces})")
        target_namespaces = namespaces if namespaces else self.list_namespaces()
        items: list[dict[str, Any]] = []
        for namespace in target_namespaces:
            if "Deployment" in resource_types:
                items.extend(self._collect_deployment_request_details(namespace=namespace))
            if "StatefulSet" in resource_types:
                items.extend(self._collect_statefulset_request_details(namespace=namespace))
        return items

    def get_workload_instance_type_from_pods(self, namespace: str, selector_match_labels: dict[str, str]) -> str | None:
        self.logger.debug(
            f"get_workload_instance_type_from_pods(namespace: {namespace}, selector_match_labels: {selector_match_labels})"
        )
        label_selector = self._build_label_selector(match_labels=selector_match_labels)
        if not label_selector:
            return None
        pods = self.core_v1.list_namespaced_pod(namespace=namespace, label_selector=label_selector)
        node_instance_type_counter: Counter[str] = Counter()
        node_cache: dict[str, str | None] = {}
        for pod in pods.items:
            node_name = pod.spec.node_name if pod.spec else None
            if not node_name:
                continue
            if node_name not in node_cache:
                node = self.core_v1.read_node(node_name)
                node_cache[node_name] = self._get_node_instance_type_label(node=node)
            instance_type = node_cache[node_name]
            if instance_type:
                node_instance_type_counter[instance_type] += 1
        if not node_instance_type_counter:
            return None
        return node_instance_type_counter.most_common(1)[0][0]

    def get_statefulset(self, name, namespace):
        self.logger.debug(f"get_statefulset(name: {name}, namespace: {namespace})")
        return self.apps_v1.read_namespaced_stateful_set(name, namespace)

    def get_image_version(self, statefulset, container_keyword):
        self.logger.debug("get_image_version(statefulset: ..., " f"container_keyword: {container_keyword})")
        containers = statefulset.spec.template.spec.containers or []
        container = next((item for item in containers if container_keyword in item.name), None) or (containers[0] if containers else None)
        if not container or not container.image:
            return "unknown"
        image = container.image
        if "@" in image:
            image = image.split("@", 1)[0]
        if ":" in image:
            return image.rsplit(":", 1)[1]
        return "unknown"

    def get_volume_claim_template_name(self, statefulset):
        self.logger.debug("get_volume_claim_template_name(statefulset: ...)")
        templates = statefulset.spec.volume_claim_templates or []
        if not templates:
            raise RuntimeError("No volumeClaimTemplates found on statefulset.")
        for template in templates:
            if template.metadata and template.metadata.name == "prometheus-thanos-db":
                return template.metadata.name
        if len(templates) == 1 and templates[0].metadata:
            return templates[0].metadata.name
        template_names = [template.metadata.name for template in templates if template.metadata and template.metadata.name]
        raise RuntimeError("Multiple volumeClaimTemplates found on statefulset: " f"{template_names}. Unable to determine the Prometheus PVC.")

    def find_pvc_name(self, claim_name, statefulset_name, namespace):
        self.logger.debug("find_pvc_name(" f"claim_name: {claim_name}, statefulset_name: {statefulset_name}, " f"namespace: {namespace})")
        prefix = f"{claim_name}-{statefulset_name}-"
        pvcs = self.core_v1.list_namespaced_persistent_volume_claim(namespace)
        matches = [pvc.metadata.name for pvc in pvcs.items if pvc.metadata.name.startswith(prefix)]
        if not matches:
            raise RuntimeError("No PVCs found for claim template " f"'{claim_name}' with prefix '{prefix}'.")
        if len(matches) > 1:
            raise RuntimeError("Multiple PVCs found for claim template " f"'{claim_name}': {matches}. " "Refine logic or specify a single PVC.")
        return matches[0]

    def get_deployments(self, namespace):
        self.logger.debug(f"get_deployments(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
            return [deployment.metadata.name for deployment in deployments.items]
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing deployments in namespace {namespace}: {e}")
            exit(1)

    def get_deployment_replicas(self, namespace):
        self.logger.debug(f"get_deployment_replicas(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
            return {deployment.metadata.name: deployment.spec.replicas for deployment in deployments.items}
        except kubernetes.client.rest.ApiException as e:
            print(f"Error fetching deployment replicas in namespace {namespace}: {e}")
            exit(1)

    def get_deployment_cost_details(self, namespace):
        self.logger.debug(f"get_deployment_cost_details(namespace: {namespace})")
        try:
            deployments = self.apps_v1.list_namespaced_deployment(namespace)
        except kubernetes.client.rest.ApiException as e:
            raise RuntimeError(f"Error fetching deployment cost details in namespace {namespace}: {e}")
            
        if not deployments.items:
            return {}
        
        # TODO: fetch cost details from aws pricing api
        per_cpu_cost = 0.0
        per_ram_cost = 0.0
        # per_cpu_cost = aws_client.get_cpu_price(service="EC2", region=region, instance_type=instance_type)
        # per_ram_cost = aws_client.get_price(service="EC2", region=region, instance_type=instance_type)
        
        cost_details = {}
        for deployment in deployments.items:
            name = deployment.metadata.name
            replicas = deployment.spec.replicas or 0
            cpu_limit = "N/A"
            ram_limit = "N/A"

            if deployment.spec.template.spec.containers:
                for container in deployment.spec.template.spec.containers:
                    if container.resources and container.resources.limits:
                        cpu_limit = container.resources.limits.get("cpu", "N/A")
                        ram_limit = container.resources.limits.get("memory", "N/A")
                        break

            # # TODO: Add cost details to the deployment
            # # TODO: fetch cost details from aws pricing api
            # cpu_cost = cpu_request * per_cpu_cost
            # ram_cost = ram_request * per_ram_cost
            # total_cost = cpu_cost + ram_cost
            cost_details[name] = {
                "replicas": replicas,
                # "cpu_request": cpu_request,
                # "ram_request": ram_request,
                "cpu_limit": cpu_limit,
                "ram_limit": ram_limit,
                # "cpu_cost": cpu_cost
                # "ram_cost": ram_cost,
                # "total_cost": total_cost
            }
        return cost_details

        
    def remove_hpa_downscaling_annotations(self, hpa_item):
        self.logger.debug(f"remove_hpa_downscaling_annotations(hpa_item: {hpa_item})")
        if not hpa_item.metadata.annotations:
            return

        annotations = hpa_item.metadata.annotations.copy()
        keys_to_remove = ["downscaler/downtime-replicas", "downscaler/uptime"]

        for key in keys_to_remove:
            annotations.pop(key, None)

        metadata = kubernetes.V1ObjectMeta(annotations=annotations)
        body = kubernetes.V2HorizontalPodAutoscaler(metadata=metadata)

        try:
            self.autoscaling_v2.patch_namespaced_horizontal_pod_autoscaler(name=hpa_item.metadata.name, namespace=hpa_item.metadata.namespace, body=body)
        except kubernetes.client.rest.ApiException as e:
            print("Error removing HPA annotations in " f"{hpa_item.metadata.namespace}/{hpa_item.metadata.name}: {e}")

    def remove_hpa_downscaling_annotations_namespaced(self, namespace):
        self.logger.debug(f"remove_hpa_downscaling_annotations_namespaced(namespace: {namespace})")
        try:
            result = self.autoscaling_v2.list_namespaced_horizontal_pod_autoscaler(namespace)
            for item in result.items:
                self.remove_hpa_downscaling_annotations(item)
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing HPAs in namespace {namespace}: {e}")

    def add_hpa_downscaling_annotations(self, hpa_item):
        self.logger.debug(f"add_hpa_downscaling_annotations(hpa_item: {hpa_item})")
        annotations = hpa_item.metadata.annotations or {}
        annotations["downscaler/downtime-replicas"] = "1"
        annotations["downscaler/uptime"] = "Mon-Fri 05:30-20:30 Europe/Berlin"

        metadata = kubernetes.V1ObjectMeta(annotations=annotations)
        body = kubernetes.V2HorizontalPodAutoscaler(metadata=metadata)

        try:
            self.autoscaling_v2.patch_namespaced_horizontal_pod_autoscaler(name=hpa_item.metadata.name, namespace=hpa_item.metadata.namespace, body=body)
        except kubernetes.client.rest.ApiException as e:
            print("Error adding HPA annotations in " f"{hpa_item.metadata.namespace}/{hpa_item.metadata.name}: {e}")

    def add_hpa_downscaling_annotations_namespaced(self, namespace):
        self.logger.debug(f"add_hpa_downscaling_annotations_namespaced(namespace: {namespace})")
        try:
            result = self.autoscaling_v2.list_namespaced_horizontal_pod_autoscaler(namespace)
            for item in result.items:
                self.add_hpa_downscaling_annotations(item)
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing HPAs in namespace {namespace}: {e}")

    def list_eso_crds(self):
        self.logger.debug("list_eso_crds()")
        try:
            crds = self.apiextensions_v1.list_custom_resource_definition()
        except kubernetes.client.rest.ApiException as e:
            print(f"Error listing CRDs: {e}")
            exit(1)
        
        eso_crds = []
        for crd in crds.items:
            group = crd.spec.group
            if not group.endswith("external-secrets.io"):
                continue

            storage_version = None
            for version in crd.spec.versions:
                if version.storage:
                    storage_version = version.name
                    break

            if not storage_version and crd.spec.versions:
                storage_version = crd.spec.versions[0].name

            eso_crds.append({"group": group, "version": storage_version, "kind": crd.spec.names.kind, "plural": crd.spec.names.plural, "scope": crd.spec.scope})

    def list_grafana_custom_resources(self, plural: str) -> list[dict[str, Any]]:
        self.logger.debug(f"list_grafana_custom_resources(plural: {plural})")
        if not plural:
            raise ValueError("Grafana custom resource plural must be defined.")
        return self.get_custom_resources_all_namespaces(group="integreatly.org", version="v1alpha1", plural=plural)

    def get_custom_resources_all_namespaces(self, group, version, plural):
        self.logger.debug("get_custom_resources_all_namespaces(" f"group: {group}, version: {version}, plural: {plural})")
        try:
            result = self.custom_objects.list_cluster_custom_object(group=group, version=version, plural=plural)
            return result.get("items", [])
        except kubernetes.client.rest.ApiException as e:
            print("Error listing custom resources for " f"{group}/{version}/{plural}: {e}")
            exit(1)

    def extract_eso_resources(self, output_dir):
        self.logger.debug(f"extract_eso_resources(output_dir: {output_dir})")
        eso_crds = self.list_eso_crds()

        for crd in eso_crds:
            if not crd["version"]:
                self.logger.warning(f"Skipping CRD with no version: {crd}")
                continue
            resources = self.get_custom_resources_all_namespaces(group=crd["group"], version=crd["version"], plural=crd["plural"])

            kind_dir = crd["kind"].lower()

            for resource in resources:
                metadata = resource.get("metadata", {})
                name = self._sanitize_filename(metadata.get("name", "unknown"))
                namespace = metadata.get("namespace") or "cluster"

                output_path = f"{output_dir}/{kind_dir}/{namespace}"
                self.file_manager.create_directory(output_path)

                file_path = f"{output_path}/{name}.yaml"
                self.file_manager.write_yaml_file(file_path, resource)

    def get_deployment(self, namespace: str, name: str):
        self.logger.debug(f"get_deployment(namespace: {namespace}, name: {name})")
        return self.apps_v1.read_namespaced_deployment(name=name, namespace=namespace)

    def is_deployment_scaled_to_zero(self, namespace: str, name: str) -> bool:
        self.logger.debug(f"is_deployment_scaled_to_zero(namespace: {namespace}, name: {name})")
        deployment = self.get_deployment(namespace=namespace, name=name)
        desired = deployment.spec.replicas or 0
        available = deployment.status.available_replicas or 0
        return desired == 0 and available == 0

    def get_flux_kustomization(self, namespace: str, name: str) -> Dict:
        self.logger.debug(f"get_flux_kustomization(namespace: {namespace}, name: {name})")
        supported_versions = ["v1", "v1beta2", "v1beta1"]
        last_error = None
        for version in supported_versions:
            try:
                return self.custom_objects.get_namespaced_custom_object(group="kustomize.toolkit.fluxcd.io", version=version, namespace=namespace, plural="kustomizations", name=name)
            except kubernetes.client.rest.ApiException as exc:
                if exc.status in (404, 400):
                    last_error = exc
                    continue
                raise
        raise RuntimeError("Could not read Flux Kustomization " f"{namespace}/{name}. Last error: {last_error}")

    def is_flux_kustomization_suspended(self, namespace: str, name: str) -> bool:
        kustomization = self.get_flux_kustomization(namespace=namespace, name=name)
        spec = kustomization.get("spec", {})
        return bool(spec.get("suspend", False))

    def list_node_names_for_node_group(self, node_group_name: str) -> List[str]:
        self.logger.debug(f"list_node_names_for_node_group(node_group_name: {node_group_name})")
        selector = f"eks.amazonaws.com/nodegroup={node_group_name}"
        nodes = self.core_v1.list_node(label_selector=selector)
        names = [item.metadata.name for item in nodes.items if item.metadata]
        names.sort()
        return names

    def wait_for_nodes_ready(self, node_names: List[str], timeout_seconds: int = 1800, poll_seconds: int = 15) -> bool:
        self.logger.debug(
            f"wait_for_nodes_ready(node_names: {node_names}, timeout_seconds: {timeout_seconds}, poll_seconds: {poll_seconds})"
        )
        deadline = time.time() + timeout_seconds
        pending = set(node_names)
        while time.time() < deadline:
            for node_name in list(pending):
                if self._is_node_ready(node_name):
                    pending.remove(node_name)
            if not pending:
                return True
            time.sleep(poll_seconds)
        raise TimeoutError("Timed out waiting for nodes to become Ready: " f"{', '.join(sorted(pending))}")

    def cordon_node(self, node_name: str) -> None:
        self.logger.debug(f"cordon_node(node_name: {node_name})")
        body = {"spec": {"unschedulable": True}}
        self.core_v1.patch_node(node_name, body)

    def uncordon_node(self, node_name: str) -> None:
        self.logger.debug(f"uncordon_node(node_name: {node_name})")
        body = {"spec": {"unschedulable": False}}
        self.core_v1.patch_node(node_name, body)

    def drain_node(self, node_name: str, timeout_seconds: int = 600) -> tuple[bool, List[str]]:
        self.logger.debug(f"drain_node(node_name: {node_name}, timeout_seconds: {timeout_seconds})")
        eviction_errors: List[str] = []
        self.cordon_node(node_name=node_name)
        self.logger.info(f"Cordoned node {node_name}.")

        pods_on_node = self.list_non_daemonset_non_static_pods_on_node(node_name=node_name)
        self.logger.info(f"Found {len(pods_on_node)} non-daemonset/non-static pods on node {node_name}.")

        evicted_pods: List[Dict[str, str]] = []

        for pod in pods_on_node:
            namespace = pod["namespace"]
            pod_name = pod["pod_name"]

            eviction = kubernetes.client.V1Eviction(metadata=kubernetes.client.V1ObjectMeta(name=pod_name, namespace=namespace))
            try:
                self._create_namespaced_pod_eviction(namespace=namespace, pod_name=pod_name, eviction=eviction)
            except kubernetes.client.rest.ApiException as exc:
                error_message = f"ERROR: Eviction failed for {namespace}/{pod_name}: {exc}"
                self.logger.warning(error_message)
                eviction_errors.append(error_message)
                continue

            evicted_pods.append(pod)
            self.logger.info(f"Eviction created for {namespace}/{pod_name}.")
            time.sleep(2)

        deadline = time.time() + timeout_seconds
        pending_pods = evicted_pods
        while pending_pods and time.time() < deadline:
            remaining: List[Dict[str, str]] = []
            for pod in pending_pods:
                current_pod = self.get_pod_if_exists(namespace=pod["namespace"], pod_name=pod["pod_name"])
                if current_pod is None:
                    continue

                if current_pod.metadata.uid != pod["pod_uid"]:
                    continue

                remaining.append(pod)

            pending_pods = remaining
            if pending_pods:
                time.sleep(3)

        if pending_pods:
            pending_names = ", ".join([f"{pod['namespace']}/{pod['pod_name']}" for pod in pending_pods])
            timeout_error = f"ERROR: Timed out waiting for pod deletions after eviction: {pending_names}"
            self.logger.warning(timeout_error)
            return False, [*eviction_errors, timeout_error]

        if eviction_errors:
            self.logger.warning("Drain completed with eviction errors.")
            return False, eviction_errors

        self.logger.info("All successfully evicted pods were deleted.")
        return True, []

    def list_multi_replicas_pods_on_node(self, node_name: str, owner_resource_types: List[str]) -> List[Dict[str, str]]:
        self.logger.debug(
            "list_multi_replicas_pods_on_node(" f"node_name: {node_name}, owner_resource_types: {owner_resource_types})"
        )
        pods = self.core_v1.list_pod_for_all_namespaces(field_selector=f"spec.nodeName={node_name}")
        multi_replicas_pods: List[Dict[str, str]] = []
        for pod in pods.items:
            owner_refs = pod.metadata.owner_references or []
            owner = next((owner_ref for owner_ref in owner_refs if owner_ref.kind in owner_resource_types), None)
            if not owner:
                continue
            multi_replicas_pods.append(
                {
                    "namespace": pod.metadata.namespace,
                    "owner_kind": owner.kind,
                    "owner_name": owner.name,
                    "pod_name": pod.metadata.name,
                    "pod_uid": pod.metadata.uid,
                }
            )
        return multi_replicas_pods

    def list_non_daemonset_non_static_pods_on_node(self, node_name: str) -> List[Dict[str, str]]:
        self.logger.debug(f"list_non_daemonset_non_static_pods_on_node(node_name: {node_name})")
        pods = self.core_v1.list_pod_for_all_namespaces(field_selector=f"spec.nodeName={node_name}")
        results: List[Dict[str, str]] = []
        for pod in pods.items:
            owner_refs = pod.metadata.owner_references or []
            if any(owner.kind == "DaemonSet" for owner in owner_refs):
                continue

            annotations = pod.metadata.annotations or {}
            if "kubernetes.io/config.mirror" in annotations:
                continue

            results.append({"namespace": pod.metadata.namespace, "pod_name": pod.metadata.name, "pod_uid": pod.metadata.uid})
        return results

    def get_pod_if_exists(self, namespace: str, pod_name: str):
        self.logger.debug(f"get_pod_if_exists(namespace: {namespace}, pod_name: {pod_name})")
        try:
            return self.core_v1.read_namespaced_pod(name=pod_name, namespace=namespace)
        except kubernetes.client.rest.ApiException as exc:
            if exc.status == 404:
                return None
            raise

    def wait_for_pod_relocated_off_node(self, namespace: str, pod_name: str, old_pod_uid: str, old_node_name: str, timeout_seconds: int = 600, poll_seconds: int = 3) -> str:
        self.logger.debug(
            f"wait_for_pod_relocated_off_node(namespace: {namespace}, pod_name: {pod_name}, old_pod_uid: {old_pod_uid}, "
            f"old_node_name: {old_node_name}, timeout_seconds: {timeout_seconds}, poll_seconds: {poll_seconds})"
        )
        deadline = time.time() + timeout_seconds
        while time.time() < deadline:
            pod = self.get_pod_if_exists(namespace=namespace, pod_name=pod_name)
            if pod is None:
                return "evicted or removed"

            if pod.metadata.uid == old_pod_uid:
                if pod.spec.node_name != old_node_name and self._is_pod_ready(pod):
                    return f"rescheduled to {pod.spec.node_name}"
                time.sleep(poll_seconds)
                continue

            if pod.spec.node_name == old_node_name:
                time.sleep(poll_seconds)
                continue

            if self._is_pod_ready(pod):
                return f"rescheduled to {pod.spec.node_name}"

            time.sleep(poll_seconds)

        raise TimeoutError(
            "Timed out waiting for pod relocation off old node: "
            f"{namespace}/{pod_name} from {old_node_name}"
        )

    def list_multi_replicas_pods(self, namespace: str, owner_resource_name: str, owner_resource_type: str) -> List:
        self.logger.debug(
            "list_multi_replicas_pods("
            f"namespace: {namespace}, owner_resource_name: {owner_resource_name}, owner_resource_type: {owner_resource_type})"
        )
        pods = self.core_v1.list_namespaced_pod(namespace=namespace)
        filtered = []
        for pod in pods.items:
            owner_refs = pod.metadata.owner_references or []
            is_owned = any(owner.kind == owner_resource_type and owner.name == owner_resource_name for owner in owner_refs)
            if is_owned:
                filtered.append(pod)
        return filtered

    def count_multi_replicas_pods_on_node(self, namespace: str, owner_resource_name: str, node_name: str, owner_resource_type: str) -> int:
        self.logger.debug(
            "count_multi_replicas_pods_on_node("
            f"namespace: {namespace}, owner_resource_name: {owner_resource_name}, node_name: {node_name}, "
            f"owner_resource_type: {owner_resource_type})"
        )
        pods = self.list_multi_replicas_pods(namespace, owner_resource_name, owner_resource_type)
        return sum(1 for pod in pods if pod.spec.node_name == node_name)

    def get_multi_replicas_unavailable_pods_count(self, namespace: str, owner_resource_name: str, owner_resource_type: str) -> int:
        self.logger.debug(
            "get_multi_replicas_unavailable_pods_count("
            f"namespace: {namespace}, owner_resource_name: {owner_resource_name}, owner_resource_type: {owner_resource_type})"
        )
        if owner_resource_type == "StatefulSet":
            resource = self.apps_v1.read_namespaced_stateful_set(owner_resource_name, namespace)
            desired = resource.spec.replicas or 0
        elif owner_resource_type == "ReplicaSet":
            resource = self.apps_v1.read_namespaced_replica_set(owner_resource_name, namespace)
            desired = resource.spec.replicas or 0
        elif owner_resource_type == "StrimziPodSet":
            resource = self.custom_objects.get_namespaced_custom_object(
                group="core.strimzi.io",
                version="v1beta2",
                namespace=namespace,
                plural="strimzipodsets",
                name=owner_resource_name,
            )
            desired = int(resource.get("spec", {}).get("replicas") or 0)
        else:
            raise ValueError(f"Unsupported owner resource type: {owner_resource_type}")
        pods = self.list_multi_replicas_pods(namespace=namespace, owner_resource_name=owner_resource_name, owner_resource_type=owner_resource_type)
        ready = sum(1 for pod in pods if self._is_pod_ready(pod))
        unavailable = desired - ready
        if unavailable < 0:
            return 0
        return unavailable

    def delete_pod(self, namespace: str, pod_name: str) -> None:
        self.logger.debug(f"delete_pod(namespace: {namespace}, pod_name: {pod_name})")
        self.core_v1.delete_namespaced_pod(name=pod_name, namespace=namespace)

    def get_pod(self, namespace: str, pod_name: str):
        self.logger.debug(f"get_pod(namespace: {namespace}, pod_name: {pod_name})")
        return self.core_v1.read_namespaced_pod(name=pod_name, namespace=namespace)

    def wait_for_recreated_pod_ready(self, namespace: str, pod_name: str, old_pod_uid: str, timeout_seconds: int = 1800, poll_seconds: int = 10) -> None:
        self.logger.debug(
            f"wait_for_recreated_pod_ready(namespace: {namespace}, pod_name: {pod_name}, old_pod_uid: {old_pod_uid}, "
            f"timeout_seconds: {timeout_seconds}, poll_seconds: {poll_seconds})"
        )
        deadline = time.time() + timeout_seconds
        while time.time() < deadline:
            pod = self.core_v1.read_namespaced_pod(name=pod_name, namespace=namespace)
            if pod.metadata.uid == old_pod_uid:
                time.sleep(poll_seconds)
                continue
            if self._is_pod_ready(pod):
                return
            time.sleep(poll_seconds)
        raise TimeoutError(f"Timed out waiting for recreated pod readiness: {namespace}/{pod_name}")

    def read_pod_logs(self, namespace: str, pod_name: str, since_seconds: int = 600, tail_lines: int = 1000) -> str:
        self.logger.debug(
            f"read_pod_logs(namespace: {namespace}, pod_name: {pod_name}, since_seconds: {since_seconds}, tail_lines: {tail_lines})"
        )
        return self.core_v1.read_namespaced_pod_log(name=pod_name, namespace=namespace, since_seconds=since_seconds, tail_lines=tail_lines)

    def pod_logs_have_errors(self, namespace: str, pod_name: str, error_regex: str, allowlist_regexes: List[str], since_seconds: int = 600) -> bool:
        self.logger.debug(
            f"pod_logs_have_errors(namespace: {namespace}, pod_name: {pod_name}, error_regex: {error_regex}, "
            f"allowlist_regexes: {allowlist_regexes}, since_seconds: {since_seconds})"
        )
        logs = self.read_pod_logs(namespace=namespace, pod_name=pod_name, since_seconds=since_seconds)
        if not logs:
            return False
        error_pattern = re.compile(error_regex)
        allowlist_patterns = [re.compile(pattern) for pattern in allowlist_regexes]
        for line in logs.splitlines():
            if not error_pattern.search(line):
                continue
            is_allowlisted = any(pattern.search(line) for pattern in allowlist_patterns)
            if not is_allowlisted:
                return True
        return False

    def list_pdbs_with_zero_disruptions_allowed(self) -> List[Dict[str, str]]:
        self.logger.debug("list_pdbs_with_zero_disruptions_allowed()")
        pdbs = self.policy_v1.list_pod_disruption_budget_for_all_namespaces()
        blocked: List[Dict[str, str]] = []
        for pdb in pdbs.items:
            disruptions_allowed = pdb.status.disruptions_allowed or 0
            if disruptions_allowed == 0:
                blocked.append({"namespace": pdb.metadata.namespace, "name": pdb.metadata.name})
        return blocked


if __name__ == "__main__":
    
    kubernetes_client = KubernetesClient(context="infrastructure")

    deployment_name = "kubernetes-client-test"
    namespace = "my_namespace"
    flux_kustomization_name = "my_flux_kustomization"
    flux_kustomization_namespace = "my_namespace"

    statefulset_name = "kubernetes-client-test"
    statefulset_namespace = "my_namespace"

    node_name = "ip-x-x-x-x.eu-central-1.compute.internal"
    print("Hello, world!")

    namespace = "ingress"
    pod_name="ingress-nginx-controller-x-x-x-x"
    # kubernetes_client.drain_node(namespace="ingress", pod_name="ingress-nginx-controller-x-x-x-x")
    eviction = kubernetes.client.V1Eviction(metadata=kubernetes.client.V1ObjectMeta(name=pod_name, namespace=namespace))
    kubernetes_client._create_namespaced_pod_eviction(namespace=namespace, pod_name=pod_name, eviction=eviction)

    # print("Getting deployment")
    # deployment = kubernetes_client.get_deployment(namespace=namespace, name=deployment_name)
    # print(deployment)
    # input("Press Enter to continue...")

    # print("Checking if deployment is scaled to zero")
    # is_deployment_scaled_to_zero = kubernetes_client.is_deployment_scaled_to_zero(namespace=namespace, name=deployment_name)
    # print(is_deployment_scaled_to_zero)
    # input("Press Enter to continue...")
    
    # print("Getting flux kustomization")
    # flux_kustomization = kubernetes_client.get_flux_kustomization(namespace=flux_kustomization_namespace, name=flux_kustomization_name)
    # print(flux_kustomization)
    # input("Press Enter to continue...")
    
    # print("Checking if flux kustomization is suspended")
    # is_flux_kustomization_suspended = kubernetes_client.is_flux_kustomization_suspended(namespace=namespace, name=flux_kustomization_name)
    # print(is_flux_kustomization_suspended)
    # input("Press Enter to continue...")



from pathlib import Path

from core.errors.exceptions import HapeOperationError, HapeValidationError
from core.errors.messages.csv_error_messages import get_csv_error_message
from core.logging import LocalLogging
from utils.csv_manager import CsvManager


class CsvService:
    def __init__(self, csv_manager: CsvManager | None = None) -> None:
        self.csv_manager = csv_manager or CsvManager()
        self.logger = LocalLogging.get_logger("hape.csv_service")

    def _read_json_rows_from_file(self, json_file: str) -> list[dict]:
        try:
            return self.csv_manager.read_json_rows(json_file)
        except ValueError as exc:
            self._raise_json_validation_error(exc)
        except Exception as exc:
            raise HapeValidationError(
                code="CSV_INVALID_JSON",
                message=get_csv_error_message("CSV_INVALID_JSON"),
            ) from exc
        raise HapeValidationError(
            code="CSV_INVALID_JSON",
            message=get_csv_error_message("CSV_INVALID_JSON"),
        )

    def _read_json_rows_from_content(self, json_content: str) -> list[dict]:
        try:
            return self.csv_manager.parse_json_rows(json_content)
        except ValueError as exc:
            self._raise_json_validation_error(exc)
        except Exception as exc:
            raise HapeValidationError(
                code="CSV_INVALID_JSON",
                message=get_csv_error_message("CSV_INVALID_JSON"),
            ) from exc
        raise HapeValidationError(
            code="CSV_INVALID_JSON",
            message=get_csv_error_message("CSV_INVALID_JSON"),
        )

    def _raise_json_validation_error(self, error: ValueError) -> None:
        message = str(error)
        if message == "JSON must be an array of objects.":
            raise HapeValidationError(
                code="CSV_JSON_ARRAY_REQUIRED",
                message=get_csv_error_message("CSV_JSON_ARRAY_REQUIRED"),
            ) from error
        if message == "JSON array is empty.":
            raise HapeValidationError(
                code="CSV_JSON_ARRAY_EMPTY",
                message=get_csv_error_message("CSV_JSON_ARRAY_EMPTY"),
            ) from error
        if message == "Each JSON array item must be an object.":
            raise HapeValidationError(
                code="CSV_JSON_ROW_OBJECT_REQUIRED",
                message=get_csv_error_message("CSV_JSON_ROW_OBJECT_REQUIRED"),
            ) from error
        raise HapeValidationError(
            code="CSV_INVALID_JSON",
            message=get_csv_error_message("CSV_INVALID_JSON"),
        ) from error

    def _validate_delimiter(self, delimiter: str) -> None:
        if len(delimiter) != 1:
            raise HapeValidationError(
                code="CSV_DELIMITER_INVALID",
                message=get_csv_error_message("CSV_DELIMITER_INVALID"),
            )

    def from_json(self, output_file: str, json_file: str | None = None, json_content: str | None = None, delimiter: str = ",") -> str:
        self.logger.debug(
            f"from_json(output_file: {output_file}, json_file: {json_file}, json_content_provided: {bool(json_content)}, delimiter: {delimiter})"
        )
        self._validate_delimiter(delimiter)
        if not output_file or not output_file.strip():
            raise HapeValidationError(
                code="CSV_OUTPUT_FILE_REQUIRED",
                message=get_csv_error_message("CSV_OUTPUT_FILE_REQUIRED"),
            )
        if bool(json_file) == bool(json_content):
            code = "CSV_JSON_SOURCE_REQUIRED" if not json_file and not json_content else "CSV_JSON_SOURCE_CONFLICT"
            raise HapeValidationError(
                code=code,
                message=get_csv_error_message(code),
            )
        if json_file:
            if not Path(json_file).exists():
                raise HapeValidationError(
                    code="CSV_JSON_FILE_NOT_FOUND",
                    message=get_csv_error_message("CSV_JSON_FILE_NOT_FOUND", json_path=json_file),
                )
            json_rows = self._read_json_rows_from_file(json_file)
        else:
            json_rows = self._read_json_rows_from_content(json_content or "")
        try:
            self.csv_manager.write_csv(output_file, json_rows, delimiter=delimiter)
        except Exception as exc:
            raise HapeOperationError(
                code="CSV_WRITE_FAILED",
                message=get_csv_error_message("CSV_WRITE_FAILED", output_path=output_file),
            ) from exc
        return output_file

    def to_json(self, csv_file: str, output_file: str, delimiter: str = ",") -> str:
        self.logger.debug(f"to_json(csv_file: {csv_file}, output_file: {output_file}, delimiter: {delimiter})")
        self._validate_delimiter(delimiter)
        if not csv_file or not csv_file.strip():
            raise HapeValidationError(
                code="CSV_INPUT_FILE_REQUIRED",
                message=get_csv_error_message("CSV_INPUT_FILE_REQUIRED"),
            )
        if not output_file or not output_file.strip():
            raise HapeValidationError(
                code="CSV_OUTPUT_FILE_REQUIRED",
                message=get_csv_error_message("CSV_OUTPUT_FILE_REQUIRED"),
            )
        if not Path(csv_file).exists():
            raise HapeValidationError(
                code="CSV_INPUT_FILE_NOT_FOUND",
                message=get_csv_error_message("CSV_INPUT_FILE_NOT_FOUND", csv_path=csv_file),
            )
        try:
            rows = self.csv_manager.read_csv(csv_file, delimiter=delimiter)
        except Exception as exc:
            raise HapeOperationError(
                code="CSV_READ_FAILED",
                message=get_csv_error_message("CSV_READ_FAILED", csv_path=csv_file),
            ) from exc
        try:
            self.csv_manager.write_json(output_file, rows)
        except Exception as exc:
            raise HapeOperationError(
                code="CSV_WRITE_JSON_FAILED",
                message=get_csv_error_message("CSV_WRITE_JSON_FAILED", output_path=output_file),
            ) from exc
        return output_file

def _demo_from_json() -> None:
    csv_service = CsvService()
    output = csv_service.from_json(
        output_file="/tmp/hape_csv_service_from_json.csv",
        json_content='[{"name":"alpha","value":1},{"name":"beta","value":2}]',
    )
    print(output)


def _demo_to_json() -> None:
    csv_service = CsvService()
    output = csv_service.to_json(
        csv_file="/tmp/hape_csv_service_from_json.csv",
        output_file="/tmp/hape_csv_service_to_json.json",
    )
    print(output)


if __name__ == "__main__":
    _demo_from_json()
    _demo_to_json()



from typing import Callable, Optional

from core.errors.exceptions import HapeError
from core.logging import LocalLogging


class ErrorHandler:
    DEFAULT_UNKNOWN_EXIT_CODE = 99
    DEFAULT_UNKNOWN_MESSAGE = "Unexpected error. Please check logs and rerun."

    @staticmethod
    def handle(exc: Exception, print_fn: Optional[Callable[[str], None]] = None) -> int:
        LocalLogging.bootstrap()
        active_logger = LocalLogging.get_logger("hape.error_handler")
        active_print = print_fn or print

        if isinstance(exc, HapeError):
            active_logger.error(f"[{exc.code}] {exc.message}")
            active_print(f"Error: {exc.message}")
            return exc.exit_code

        active_logger.error(f"[HAPE_UNKNOWN_ERROR] {exc.__class__.__name__}: {exc}")
        active_print(f"Error: {ErrorHandler.DEFAULT_UNKNOWN_MESSAGE}")
        return ErrorHandler.DEFAULT_UNKNOWN_EXIT_CODE


if __name__ == "__main__":
    exit_code = ErrorHandler.handle(Exception("example"))
    print(f"example-exit-code={exit_code}")



from urllib.parse import urlparse


class UrlsUtils:
    @staticmethod
    def normalize_grafana_base_url(raw_base_url: str) -> str:
        if not raw_base_url or not raw_base_url.strip():
            raise ValueError("Grafana URL is required.")
        trimmed_base_url = raw_base_url.strip()
        parsed_url = urlparse(trimmed_base_url)
        if parsed_url.scheme and parsed_url.netloc:
            return f"{parsed_url.scheme}://{parsed_url.netloc}"
        raise ValueError(f"Grafana URL must include scheme and host. Value is '{raw_base_url}'.")

    @staticmethod
    def normalize_atlassian_base_url(raw_domain: str) -> str:
        if not raw_domain:
            raise ValueError("ATLASSIAN_DOMAIN is required.")
        raw_domain = raw_domain.strip()
        if not raw_domain:
            raise ValueError("ATLASSIAN_DOMAIN is required.")

        if raw_domain.startswith(("http://", "https://")):
            parsed = urlparse(raw_domain)
            if not parsed.netloc:
                raise ValueError("ATLASSIAN_DOMAIN must include a host.")
            return f"{parsed.scheme}://{parsed.netloc}"

        host = raw_domain.split("/", 1)[0]
        return f"https://{host}"


if __name__ == "__main__":
    print(UrlsUtils.normalize_grafana_base_url("https://grafana.example.com/path"))



import csv
import tempfile
from datetime import datetime, timezone
from pathlib import Path

import boto3
import kubernetes.config

from clients.aws_client import AwsClient
from clients.kubernetes_client import KubernetesClient
from core.errors.exceptions import HapeExternalError, HapeOperationError, HapeValidationError
from core.errors.messages.eks_deployment_cost_error_messages import get_eks_deployment_cost_error_message
from core.logging import LocalLogging
from utils.csv_manager import CsvManager
from utils.file_manager import FileManager


class EksDeploymentCostService:
    ALLOWED_RESOURCE_TYPES = ["Deployment", "StatefulSet"]
    HOURS_PER_DAY = 24
    HOURS_PER_MONTH = 24 * 30
    HOURS_PER_YEAR = 24 * 365
    DETAILS_CSV_COLUMNS = [
        "resource_type",
        "namespace",
        "name",
        "instance_type",
        "replicas",
        "cpu_request_cores_per_pod",
        "memory_request_gib_per_pod",
        "total_cpu_request_cores",
        "total_memory_request_gib",
        "usd_per_vcpu_hour",
        "usd_per_gib_hour",
        "hourly_cpu_cost_usd",
        "hourly_memory_cost_usd",
        "max_hourly_cost_usd",
        "daily_cost_usd",
        "monthly_cost_usd",
        "yearly_cost_usd",
    ]

    def __init__(self, csv_manager: CsvManager | None = None, file_manager: FileManager | None = None) -> None:
        self.csv_manager = csv_manager or CsvManager()
        self.file_manager = file_manager or FileManager()
        self.logger = LocalLogging.get_logger("hape.eks_deployment_cost_service")

    @staticmethod
    def _normalize_csv_argument(value: str | None) -> list[str] | None:
        if value is None:
            return None
        raw_items = [item.strip() for item in value.split(",")]
        items = [item for item in raw_items if item]
        return items or None

    def _validate_inputs(self, kube_context: str, output_dir: str, top_n: int, resource_types: list[str], use_incluster_config: bool) -> None:
        if not use_incluster_config and (not kube_context or not kube_context.strip()):
            raise HapeValidationError(code="EDC_KUBE_CONTEXT_REQUIRED", message=get_eks_deployment_cost_error_message("EDC_KUBE_CONTEXT_REQUIRED"))
        if not output_dir or not output_dir.strip():
            raise HapeValidationError(code="EDC_OUTPUT_DIR_REQUIRED", message=get_eks_deployment_cost_error_message("EDC_OUTPUT_DIR_REQUIRED"))
        if top_n <= 0:
            raise HapeValidationError(code="EDC_TOP_N_INVALID", message=get_eks_deployment_cost_error_message("EDC_TOP_N_INVALID"))
        invalid_resource_types = [item for item in resource_types if item not in self.ALLOWED_RESOURCE_TYPES]
        if invalid_resource_types:
            raise HapeValidationError(code="EDC_RESOURCE_TYPES_INVALID", message=get_eks_deployment_cost_error_message("EDC_RESOURCE_TYPES_INVALID", resource_types=", ".join(invalid_resource_types), allowed_resource_types=", ".join(self.ALLOWED_RESOURCE_TYPES)))

    @staticmethod
    def _round_currency(value: float) -> float:
        return round(value, 6)

    def _build_workload_cost_rows(self, workloads: list[dict], pricing_by_instance_type: dict[str, dict]) -> list[dict]:
        rows: list[dict] = []
        for workload in workloads:
            instance_type = workload["instance_type"]
            pricing_details = pricing_by_instance_type[instance_type]
            hourly_instance_price_usd = float(pricing_details["hourly_instance_price_usd"])
            instance_vcpu = float(pricing_details["vcpu"])
            instance_memory_gib = float(pricing_details["memory_gib"])
            usd_per_vcpu_hour = hourly_instance_price_usd / instance_vcpu
            usd_per_gib_hour = hourly_instance_price_usd / instance_memory_gib
            replicas = workload.get("replicas")
            replicas_effective = replicas or 1
            cpu_request_cores_per_pod = float(workload.get("cpu_request_cores_per_pod", 0.0))
            memory_request_gib_per_pod = float(workload.get("memory_request_gib_per_pod", 0.0))
            total_cpu_request_cores = cpu_request_cores_per_pod * replicas_effective
            total_memory_request_gib = memory_request_gib_per_pod * replicas_effective
            hourly_cpu_cost_usd = self._round_currency(total_cpu_request_cores * usd_per_vcpu_hour)
            hourly_memory_cost_usd = self._round_currency(total_memory_request_gib * usd_per_gib_hour)
            max_hourly_cost_usd = self._round_currency(max(hourly_cpu_cost_usd, hourly_memory_cost_usd))
            row = {
                "resource_type": workload.get("resource_type"),
                "namespace": workload.get("namespace"),
                "name": workload.get("name"),
                "instance_type": instance_type,
                "replicas": replicas,
                "cpu_request_cores_per_pod": cpu_request_cores_per_pod,
                "memory_request_gib_per_pod": memory_request_gib_per_pod,
                "total_cpu_request_cores": total_cpu_request_cores,
                "total_memory_request_gib": total_memory_request_gib,
                "usd_per_vcpu_hour": usd_per_vcpu_hour,
                "usd_per_gib_hour": usd_per_gib_hour,
                "hourly_cpu_cost_usd": hourly_cpu_cost_usd,
                "hourly_memory_cost_usd": hourly_memory_cost_usd,
                "max_hourly_cost_usd": max_hourly_cost_usd,
                "daily_cost_usd": self._round_currency(max_hourly_cost_usd * self.HOURS_PER_DAY),
                "monthly_cost_usd": self._round_currency(max_hourly_cost_usd * self.HOURS_PER_MONTH),
                "yearly_cost_usd": self._round_currency(max_hourly_cost_usd * self.HOURS_PER_YEAR),
            }
            rows.append(row)
        return rows

    @staticmethod
    def _sum_cost(rows: list[dict], key: str) -> float:
        return round(sum(float(row.get(key, 0.0)) for row in rows), 6)

    def _build_summary_report(
        self,
        kube_context: str,
        aws_region: str,
        namespaces: list[str] | None,
        ignored_namespaces: list[str],
        resource_types: list[str],
        top_n: int,
        pricing_by_instance_type: dict[str, dict],
        rows_sorted: list[dict],
    ) -> dict:
        top_rows = rows_sorted[:top_n]
        pricing_summary = {}
        for instance_type, pricing_details in pricing_by_instance_type.items():
            pricing_summary[instance_type] = {
                "hourly_instance_price_usd": pricing_details["hourly_instance_price_usd"],
                "vcpu": pricing_details["vcpu"],
                "memory_gib": pricing_details["memory_gib"],
                "usd_per_vcpu_hour": pricing_details["hourly_instance_price_usd"] / pricing_details["vcpu"],
                "usd_per_gib_hour": pricing_details["hourly_instance_price_usd"] / pricing_details["memory_gib"],
            }
        return {
            "metadata": {
                "generated_at_utc": datetime.now(timezone.utc).isoformat(),
                "kube_context": kube_context,
                "aws_region": aws_region,
                "namespaces": namespaces if namespaces is not None else ["*"],
                "ignored_namespaces": ignored_namespaces,
                "resource_types": resource_types,
                "top_n": top_n,
            },
            "summary": {
                "workload_count": len(rows_sorted),
                "total_cpu_request_cores": self._sum_cost(rows_sorted, "total_cpu_request_cores"),
                "total_memory_request_gib": self._sum_cost(rows_sorted, "total_memory_request_gib"),
                "hourly_cpu_cost_usd": self._sum_cost(rows_sorted, "hourly_cpu_cost_usd"),
                "hourly_memory_cost_usd": self._sum_cost(rows_sorted, "hourly_memory_cost_usd"),
                "max_hourly_cost_usd": self._sum_cost(rows_sorted, "max_hourly_cost_usd"),
                "daily_cost_usd": self._sum_cost(rows_sorted, "daily_cost_usd"),
                "monthly_cost_usd": self._sum_cost(rows_sorted, "monthly_cost_usd"),
                "yearly_cost_usd": self._sum_cost(rows_sorted, "yearly_cost_usd"),
            },
            "top_costing_workloads": top_rows,
            "pricing_by_instance_type": pricing_summary
        }

    def _write_details_csv(self, output_csv_path: str, rows_sorted: list[dict]) -> None:
        normalized_rows = []
        for row in rows_sorted:
            normalized_row = {}
            for column in self.DETAILS_CSV_COLUMNS:
                normalized_row[column] = row.get(column, "")
            normalized_rows.append(normalized_row)
        if normalized_rows:
            self.csv_manager.write_csv(output_csv_path, normalized_rows)
            return
        with open(output_csv_path, "w", encoding="utf-8", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=self.DETAILS_CSV_COLUMNS)
            writer.writeheader()

    @staticmethod
    def _resolve_kube_context(kube_context: str | None, kube_config_file: str | None, use_incluster_config: bool) -> str:
        if use_incluster_config:
            return ""
        if kube_context and kube_context.strip():
            return kube_context.strip()
        try:
            contexts, active_context = kubernetes.config.list_kube_config_contexts(config_file=kube_config_file)
        except Exception as exc:
            raise HapeValidationError(
                code="EDC_KUBE_CONTEXT_REQUIRED",
                message=get_eks_deployment_cost_error_message("EDC_KUBE_CONTEXT_REQUIRED"),
            ) from exc
        if active_context and active_context.get("name"):
            return str(active_context["name"])
        if contexts:
            first_context_name = contexts[0].get("name")
            if first_context_name:
                return str(first_context_name)
        raise HapeValidationError(
            code="EDC_KUBE_CONTEXT_REQUIRED",
            message=get_eks_deployment_cost_error_message("EDC_KUBE_CONTEXT_REQUIRED"),
        )

    @staticmethod
    def _resolve_aws_profile(aws_profile: str | None) -> str | None:
        if aws_profile and aws_profile.strip():
            return aws_profile.strip()
        session = boto3.Session()
        if session.profile_name:
            return session.profile_name
        return None

    @staticmethod
    def _resolve_output_dir(output_dir: str | None) -> str:
        if output_dir and output_dir.strip():
            return output_dir.strip()
        return tempfile.mkdtemp(prefix="hape-eks-deployment-cost-")

    def generate_report(self, kube_context: str | None = None, aws_profile: str | None = None, output_dir: str | None = None, top_n: int = 20, resource_types_csv: str | None = None, namespaces_csv: str | None = None, ignored_namespaces_csv: str | None = None, aws_region: str | None = None, kube_config_file: str | None = None, use_incluster_config: bool = False) -> dict:
        self.logger.debug(
            "generate_report("
            f"kube_context: {kube_context}, aws_profile: {aws_profile}, output_dir: {output_dir}, "
            f"top_n: {top_n}, resource_types_csv: {resource_types_csv}, namespaces_csv: {namespaces_csv}, "
            f"ignored_namespaces_csv: {ignored_namespaces_csv}, aws_region: {aws_region}, kube_config_file: {kube_config_file}, "
            f"use_incluster_config: {use_incluster_config})"
        )
        resolved_kube_context = self._resolve_kube_context(
            kube_context=kube_context,
            kube_config_file=kube_config_file,
            use_incluster_config=use_incluster_config,
        )
        resolved_aws_profile = self._resolve_aws_profile(aws_profile=aws_profile)
        resolved_output_dir = self._resolve_output_dir(output_dir=output_dir)
        resource_types = self._normalize_csv_argument(resource_types_csv) or list(self.ALLOWED_RESOURCE_TYPES)
        namespaces = self._normalize_csv_argument(namespaces_csv)
        ignored_namespaces = self._normalize_csv_argument(ignored_namespaces_csv) or []
        self._validate_inputs(
            kube_context=resolved_kube_context,
            output_dir=resolved_output_dir,
            top_n=top_n,
            resource_types=resource_types,
            use_incluster_config=use_incluster_config,
        )
        kubernetes_client = KubernetesClient(
            context=resolved_kube_context,
            config_file=kube_config_file,
            use_incluster_config=use_incluster_config,
        )
        aws_client = AwsClient(profile_name=resolved_aws_profile)
        effective_region = aws_region or aws_client.get_region_name()
        ignored_namespaces_set = set(ignored_namespaces)
        effective_namespaces = namespaces or kubernetes_client.list_namespaces()
        effective_namespaces = [namespace for namespace in effective_namespaces if namespace not in ignored_namespaces_set]
        try:
            workloads = kubernetes_client.list_replica_workload_request_details(
                resource_types=resource_types,
                namespaces=effective_namespaces,
            )
        except Exception as exc:
            raise HapeExternalError(
                code="EDC_KUBERNETES_READ_FAILED",
                message=get_eks_deployment_cost_error_message("EDC_KUBERNETES_READ_FAILED"),
            ) from exc
        pricing_by_instance_type: dict[str, dict] = {}
        for workload in workloads:
            instance_type = kubernetes_client.get_workload_instance_type_from_pods(
                namespace=workload["namespace"],
                selector_match_labels=workload.get("selector_match_labels", {}),
            )
            if not instance_type:
                raise HapeExternalError(
                    code="EDC_INSTANCE_TYPE_DERIVE_FAILED",
                    message=get_eks_deployment_cost_error_message(
                        "EDC_INSTANCE_TYPE_DERIVE_FAILED",
                        resource_type=workload["resource_type"],
                        namespace=workload["namespace"],
                        name=workload["name"],
                    ),
                )
            workload["instance_type"] = instance_type
            if instance_type in pricing_by_instance_type:
                continue
            try:
                pricing_by_instance_type[instance_type] = aws_client.get_ec2_instance_type_pricing_details(
                    instance_type=instance_type,
                    region_code=effective_region,
                )
            except Exception as exc:
                raise HapeExternalError(
                    code="EDC_AWS_PRICING_LOOKUP_FAILED",
                    message=get_eks_deployment_cost_error_message(
                        "EDC_AWS_PRICING_LOOKUP_FAILED",
                        instance_type=instance_type,
                        region=effective_region,
                    ),
                ) from exc
        rows = self._build_workload_cost_rows(workloads=workloads, pricing_by_instance_type=pricing_by_instance_type)
        rows_sorted = sorted(rows, key=lambda item: item["max_hourly_cost_usd"], reverse=True)
        summary = self._build_summary_report(
            kube_context=resolved_kube_context,
            aws_region=effective_region,
            namespaces=effective_namespaces,
            ignored_namespaces=ignored_namespaces,
            resource_types=resource_types,
            top_n=top_n,
            pricing_by_instance_type=pricing_by_instance_type,
            rows_sorted=rows_sorted,
        )
        output_path = Path(resolved_output_dir)
        summary_json_path = str(output_path / "eks-deployment-cost-summary.json")
        details_csv_path = str(output_path / "eks-deployment-cost-details.csv")
        try:
            self.file_manager.create_directory(resolved_output_dir)
            self.file_manager.write_json_file(summary_json_path, summary)
            self._write_details_csv(details_csv_path, rows_sorted)
        except Exception as exc:
            raise HapeOperationError(
                code="EDC_REPORT_WRITE_FAILED",
                message=get_eks_deployment_cost_error_message("EDC_REPORT_WRITE_FAILED", output_dir=resolved_output_dir),
            ) from exc
        return {
            "summary_json_path": summary_json_path,
            "details_csv_path": details_csv_path,
        }


if __name__ == "__main__":
    eks_deployment_cost_service = EksDeploymentCostService()
    try:
        eks_deployment_cost_service.generate_report(
            kube_context="",
            aws_profile="",
            output_dir="",
        )
    except HapeValidationError as exc:
        print(exc.code)



import os
from typing import Any, Dict, Optional

from dotenv import dotenv_values

from core.config import Config
from core.errors.exceptions import HapeOperationError, HapeValidationError
from core.errors.messages.config_error_messages import get_config_error_message
from utils.file_manager import FileManager


class ConfigService:
    def __init__(self) -> None:
        self.file_manager = FileManager()
        
    @staticmethod
    def _get_parent_dir(path: str) -> str:
        parent_dir = os.path.dirname(path)
        return parent_dir or "."

    def _build_config_from_env(self, dot_env_file: Optional[str] = None) -> Dict[str, Any]:
        if dot_env_file:
            env_values = dotenv_values(dot_env_file)
            if not env_values:
                raise HapeValidationError(
                    code="CONFIG_ENV_FILE_INVALID",
                    message=get_config_error_message(
                        "CONFIG_ENV_FILE_INVALID",
                        dot_env_file=dot_env_file,
                    ),
                )
        else:
            env_values = dict(os.environ)

        config_data: Dict[str, Any] = {}
        for key in Config.get_supported_config_keys():
            value = env_values.get(key)
            if value in (None, ""):
                continue
            if key in Config.get_int_config_keys():
                try:
                    config_data[key] = int(value)
                except ValueError as exc:
                    raise HapeValidationError(
                        code="CONFIG_ENV_INT_REQUIRED",
                        message=get_config_error_message(
                            "CONFIG_ENV_INT_REQUIRED",
                            config_key=key,
                        ),
                    ) from exc
            else:
                config_data[key] = value
        return config_data
        
    def init_config_file(self, config_path: Optional[str] = None, dot_env_file: Optional[str] = None) -> str:
        resolved_path = config_path or Config.get_config_path()
        parent_dir = self._get_parent_dir(resolved_path)
        try:
            self.file_manager.create_directory(parent_dir)
        except PermissionError as exc:
            raise HapeOperationError(
                code="CONFIG_PERMISSION_DENIED",
                message=get_config_error_message(
                    "CONFIG_PERMISSION_DENIED",
                    parent_dir=parent_dir,
                ),
            ) from exc
        config_data = self._build_config_from_env(dot_env_file=dot_env_file)
        self.file_manager.write_json_file(resolved_path, config_data)
        return resolved_path



"""CLI commands package."""

__all__ = [
    "confluence_commands",
    "csv_commands",
    "eks_deployment_cost_commands",
    "eks_pre_upgrade_commands",
    "grafana_commands",
    "gitlab_commands",
    "jira_commands",
    "markdown_commands",
]



import re
from pathlib import Path


class MarkdownManager:
    _TABLE_SEPARATOR_PATTERN = re.compile(r"^\s*\|(?:\s*:?-{3,}:?\s*\|)+\s*$")

    @staticmethod
    def _is_pipe_row(line: str) -> bool:
        trimmed = line.strip()
        return trimmed.startswith("|") and trimmed.endswith("|") and "|" in trimmed[1:-1]

    @staticmethod
    def _split_pipe_row(line: str) -> list[str]:
        trimmed = line.strip().strip("|")
        return [cell.strip() for cell in trimmed.split("|")]

    def extract_tables(self, markdown_text: str) -> list[tuple[list[str], list[list[str]]]]:
        lines = markdown_text.splitlines()
        tables: list[tuple[list[str], list[list[str]]]] = []
        index = 0
        while index < len(lines):
            current_line = lines[index]
            next_index = index + 1
            if next_index >= len(lines):
                break
            if not self._is_pipe_row(current_line):
                index += 1
                continue
            if not self._TABLE_SEPARATOR_PATTERN.match(lines[next_index]):
                index += 1
                continue

            headers = self._split_pipe_row(current_line)
            body_rows: list[list[str]] = []
            index += 2
            while index < len(lines) and self._is_pipe_row(lines[index]):
                body_rows.append(self._split_pipe_row(lines[index]))
                index += 1
            tables.append((headers, body_rows))
        return tables

    def table_to_dict_rows(self, headers: list[str], body_rows: list[list[str]]) -> list[dict[str, str]]:
        rows: list[dict[str, str]] = []
        for row in body_rows:
            normalized_cells = row + ([""] * (len(headers) - len(row)))
            row_dict = {headers[i]: normalized_cells[i] for i in range(len(headers))}
            rows.append(row_dict)
        return rows

    def render_table(self, rows: list[dict[str, str]]) -> str:
        if not rows:
            raise ValueError("Cannot render markdown table from empty rows.")
        headers = list(rows[0].keys())
        header_row = f"| {' | '.join(headers)} |"
        separator_row = f"| {' | '.join(['---'] * len(headers))} |"
        body_lines = []
        for row in rows:
            cells = [self._escape_table_cell(str(row.get(header, ""))) for header in headers]
            body_lines.append(f"| {' | '.join(cells)} |")
        return "\n".join([header_row, separator_row, *body_lines])

    @staticmethod
    def _escape_table_cell(value: str) -> str:
        return value.replace("|", r"\\")

    def append_table_block(self, markdown_path: str, heading: str, table_text: str) -> None:
        path = Path(markdown_path)
        existing_content = ""
        if path.exists():
            existing_content = path.read_text(encoding="utf-8")
        block = f"{heading}\n\n{table_text}\n"
        if existing_content.strip():
            normalized = existing_content.rstrip() + "\n\n" + block
        else:
            normalized = block
        path.write_text(normalized, encoding="utf-8")


def _demo_extract_tables() -> None:
    markdown_manager = MarkdownManager()
    sample = "\n".join(
        [
            "# Demo",
            "",
            "| name | value |",
            "| --- | --- |",
            "| alpha | 1 |",
            "| beta | 2 |",
        ]
    )
    print(markdown_manager.extract_tables(sample))


def _demo_render_table() -> None:
    markdown_manager = MarkdownManager()
    rows = [{"name": "alpha", "value": "1"}, {"name": "beta", "value": "2"}]
    print(markdown_manager.render_table(rows))


if __name__ == "__main__":
    _demo_extract_tables()
    _demo_render_table()



ERROR_MESSAGES = {
    "EDC_KUBE_CONTEXT_REQUIRED": "Kubernetes context is required.",
    "EDC_AWS_PROFILE_REQUIRED": "AWS profile is required.",
    "EDC_OUTPUT_DIR_REQUIRED": "Output directory is required.",
    "EDC_TOP_N_INVALID": "Top N must be a positive integer.",
    "EDC_RESOURCE_TYPES_INVALID": "Unsupported resource types: {resource_types}. Allowed values are: {allowed_resource_types}.",
    "EDC_NAMESPACES_INVALID": "Namespaces flag format is invalid.",
    "EDC_KUBERNETES_READ_FAILED": "Failed to read Kubernetes workloads for cost report.",
    "EDC_INSTANCE_TYPE_DERIVE_FAILED": "Failed to derive EC2 instance type from running pods metadata for {resource_type} '{namespace}/{name}'.",
    "EDC_AWS_PRICING_LOOKUP_FAILED": "Failed to lookup AWS pricing details for instance type '{instance_type}' in region '{region}'.",
    "EDC_REPORT_WRITE_FAILED": "Failed to write EKS deployment cost report outputs to '{output_dir}'.",
}


def get_eks_deployment_cost_error_message(message_key: str, **kwargs: str) -> str:
    template = ERROR_MESSAGES.get(message_key, "Unknown EKS deployment cost error.")
    return template.format(**kwargs)


if __name__ == "__main__":
    print(get_eks_deployment_cost_error_message("EDC_TOP_N_INVALID"))



"""Domain error message catalogs."""

from core.errors.messages.config_error_messages import get_config_error_message
from core.errors.messages.confluence_error_messages import get_confluence_error_message
from core.errors.messages.gitlab_error_messages import get_gitlab_error_message
from core.errors.messages.jira_error_messages import get_jira_error_message
from core.errors.messages.csv_error_messages import get_csv_error_message
from core.errors.messages.eks_deployment_cost_error_messages import get_eks_deployment_cost_error_message
from core.errors.messages.markdown_error_messages import get_markdown_error_message

__all__ = [
    "get_config_error_message",
    "get_confluence_error_message",
    "get_gitlab_error_message",
    "get_jira_error_message",
    "get_csv_error_message",
    "get_eks_deployment_cost_error_message",
    "get_markdown_error_message",
]


if __name__ == "__main__":
    print(get_config_error_message("CONFIG_ENV_INT_REQUIRED"))



import html
import re

from typing import Optional

import markdown


class ConfluenceDocFormatter:
    _FENCED_BLOCK_PATTERN = re.compile(
        r"```(?P<lang>[^\n`]*)\r?\n(?P<code>.*?)\r?\n```",
        re.DOTALL,
    )

    @classmethod
    def format_markdown_to_storage(cls, markdown_text: str) -> str:
        normalized = cls._replace_fenced_blocks(markdown_text)
        return markdown.markdown(
            normalized,
            extensions=["fenced_code", "tables"],
        )

    @classmethod
    def _replace_fenced_blocks(cls, markdown_text: str) -> str:
        def replace_block(match: re.Match) -> str:
            language = match.group("lang").strip()
            code = match.group("code")
            if language.lower() == "mermaid":
                return cls._mermaid_macro(code)
            return cls._code_macro(code, language or None)

        return cls._FENCED_BLOCK_PATTERN.sub(replace_block, markdown_text)

    @classmethod
    def _mermaid_macro(cls, code: str) -> str:
        return (
            '<ac:structured-macro ac:name="mermaid">'
            "<ac:plain-text-body><![CDATA["
            f"{cls._wrap_cdata(code)}"
            "]]></ac:plain-text-body>"
            "</ac:structured-macro>"
        )

    @classmethod
    def _code_macro(cls, code: str, language: Optional[str]) -> str:
        language_parameter = ""
        if language:
            safe_language = html.escape(language)
            language_parameter = (
                f'<ac:parameter ac:name="language">{safe_language}</ac:parameter>'
            )
        return (
            '<ac:structured-macro ac:name="code">'
            f"{language_parameter}"
            "<ac:plain-text-body><![CDATA["
            f"{cls._wrap_cdata(code)}"
            "]]></ac:plain-text-body>"
            "</ac:structured-macro>"
        )

    @staticmethod
    def _wrap_cdata(text: str) -> str:
        return text.replace("]]>", "]]]]><![CDATA[>")



######################
######################
# Strict rule:
# You must not include core/logging.py to avoid circular import.
# Config class uses loggers
######################
import json
import os
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv

from utils.validation_utils import ValidationUtils


class Config:
    _config_loaded = False
    _config_data = None
    _config_path = None
    _dotenv_loaded = False
    default_config_path = os.path.expanduser("~/.hape/config.json")
    default_dotenv_path = Path(__file__).resolve().parents[1] / ".env"

    supported_config_keys = [
        "GITLAB_DOMAIN",
        "GITLAB_TOKEN",
        "GITLAB_DEFAULT_GROUP_ID",
        "ATLASSIAN_DOMAIN",
        "ATLASSIAN_EMAIL",
        "ATLASSIAN_API_KEY",
        "CONFLUENCE_CHANGELOG_PARENT_PAGE_ID",
        "CONFLUENCE_CHANGELOG_ENTRY_PAGE_TEMPLATE_ID",
        "CONFLUENCE_TEST_PARENT_PAGE_ID",
        "HAPE_EXPORTER_HOST",
        "HAPE_EXPORTER_PORT",
        "HAPE_EXPORTER_REFRESH_SECONDS",
        "HAPE_EDC_KUBE_CONTEXT",
        "HAPE_EDC_AWS_PROFILE",
        "HAPE_EDC_IGNORED_NAMESPACES",
    ]

    int_config_keys = [
        "GITLAB_DEFAULT_GROUP_ID",
        "CONFLUENCE_CHANGELOG_PARENT_PAGE_ID",
        "CONFLUENCE_CHANGELOG_ENTRY_PAGE_TEMPLATE_ID",
        "CONFLUENCE_TEST_PARENT_PAGE_ID",
        "HAPE_EXPORTER_PORT",
        "HAPE_EXPORTER_REFRESH_SECONDS",
    ]

    @staticmethod
    def _load_config() -> None:
        if Config._config_loaded:
            return
        config_path = Config.get_config_path()
        if not os.path.exists(config_path):
            raise ValueError(
                f"Config file is required. Value is not set. Path: {config_path}"
            )
        with open(config_path, "r", encoding="utf-8") as config_file:
            Config._config_data = json.load(config_file)
        Config._config_loaded = True

    @staticmethod
    def _load_dotenv() -> None:
        if Config._dotenv_loaded:
            return
        if Config.default_dotenv_path.exists():
            load_dotenv(dotenv_path=Config.default_dotenv_path, override=False)
        Config._dotenv_loaded = True

    @staticmethod
    def _get_env_value(config_key: str) -> Optional[str]:
        Config._load_dotenv()
        env_value = os.getenv(config_key)
        if env_value not in (None, ""):
            return env_value
        return None

    @staticmethod
    def _get_config_value(config_key: str, required: bool = True) -> Optional[str]:
        env_value = Config._get_env_value(config_key)
        if env_value is not None:
            return env_value
        Config._load_config()
        if Config._config_data is None:
            raise ValueError("Config file is required. Value is not set.")
        config_value = Config._config_data.get(config_key)
        if config_value not in (None, ""):
            return config_value
        if required:
            raise ValueError(
                f"{config_key} config value is required. Value is not set."
            )
        return None

    @staticmethod
    def _get_config_int(config_key: str, required: bool = True) -> Optional[int]:
        config_value = Config._get_config_value(config_key, required=required)
        if config_value is None:
            return None
        try:
            return int(config_value)
        except ValueError as exc:
            raise ValueError(
                f"{config_key} must be an integer. Value is '{config_value}'."
            ) from exc

    @staticmethod
    def _get_optional_config_value(config_key: str) -> Optional[str]:
        env_value = Config._get_env_value(config_key)
        if env_value is not None:
            return env_value
        config_path = Config.get_config_path()
        if not os.path.exists(config_path):
            return None
        Config._load_config()
        if Config._config_data is None:
            return None
        config_value = Config._config_data.get(config_key)
        if config_value in (None, ""):
            return None
        return str(config_value)

    @staticmethod
    def _get_config_value_with_default(config_key: str, default_value: str) -> str:
        config_value = Config._get_optional_config_value(config_key)
        if config_value is None:
            return default_value
        return config_value

    @staticmethod
    def _get_config_int_with_default(config_key: str, default_value: int) -> int:
        config_value = Config._get_optional_config_value(config_key)
        if config_value is None:
            return default_value
        try:
            return int(config_value)
        except ValueError as exc:
            raise ValueError(
                f"{config_key} must be an integer. Value is '{config_value}'."
            ) from exc

    @staticmethod
    def get_supported_config_keys() -> list[str]:
        return Config.supported_config_keys

    @staticmethod
    def get_int_config_keys() -> list[str]:
        return Config.int_config_keys

    @staticmethod
    def set_config_path(config_path: Optional[str]) -> None:
        if config_path:
            Config._config_path = config_path
            Config._config_loaded = False
            Config._config_data = None

    @staticmethod
    def get_config_path() -> str:
        return Config._config_path or Config.default_config_path

    @staticmethod
    def get_gitlab_token() -> str:
        token = Config._get_config_value("GITLAB_TOKEN")
        token = ValidationUtils.require_string("GITLAB_TOKEN", token)
        ValidationUtils.validate_min_length_no_spaces("GITLAB_TOKEN", token, min_length=10)
        return token

    @staticmethod
    def get_gitlab_domain() -> str:
        domain = Config._get_config_value("GITLAB_DOMAIN")
        domain = ValidationUtils.require_string("GITLAB_DOMAIN", domain)
        ValidationUtils.validate_domain("GITLAB_DOMAIN", domain)
        return domain

    @staticmethod
    def get_gitlab_default_group_id() -> int:
        group_id = Config._get_config_int("GITLAB_DEFAULT_GROUP_ID")
        ValidationUtils.validate_positive_int("GITLAB_DEFAULT_GROUP_ID", group_id)
        return group_id

    @staticmethod
    def get_atlassian_domain() -> str:
        domain = Config._get_config_value("ATLASSIAN_DOMAIN")
        domain = ValidationUtils.require_string("ATLASSIAN_DOMAIN", domain)
        ValidationUtils.validate_domain("ATLASSIAN_DOMAIN", domain)
        return domain

    @staticmethod
    def get_atlassian_email() -> str:
        email = Config._get_config_value("ATLASSIAN_EMAIL")
        email = ValidationUtils.require_string("ATLASSIAN_EMAIL", email)
        ValidationUtils.validate_email("ATLASSIAN_EMAIL", email)
        return email

    @staticmethod
    def get_atlassian_api_key() -> str:
        api_key = Config._get_config_value("ATLASSIAN_API_KEY")
        api_key = ValidationUtils.require_string("ATLASSIAN_API_KEY", api_key)
        ValidationUtils.validate_min_length_no_spaces(
            "ATLASSIAN_API_KEY",
            api_key,
            min_length=10,
        )
        return api_key

    @staticmethod
    def get_changelog_parent_page_id() -> int:
        parent_id = Config._get_config_int("CONFLUENCE_CHANGELOG_PARENT_PAGE_ID")
        ValidationUtils.validate_positive_int(
            "CONFLUENCE_CHANGELOG_PARENT_PAGE_ID",
            parent_id,
        )
        return parent_id

    @staticmethod
    def get_changelog_entry_page_template_id() -> int:
        template_id = Config._get_config_int("CONFLUENCE_CHANGELOG_ENTRY_PAGE_TEMPLATE_ID")
        ValidationUtils.validate_positive_int(
            "CONFLUENCE_CHANGELOG_ENTRY_PAGE_TEMPLATE_ID",
            template_id,
        )
        return template_id

    @staticmethod
    def get_test_parent_page_id() -> int:
        parent_id = Config._get_config_int("CONFLUENCE_TEST_PARENT_PAGE_ID")
        ValidationUtils.validate_positive_int("CONFLUENCE_TEST_PARENT_PAGE_ID", parent_id)
        return parent_id

    @staticmethod
    def get_log_level() -> str:
        log_level_raw = os.getenv("HAPE_LOG_LEVEL", "DEBUG")
        log_level = log_level_raw.strip().upper()
        allowed_levels = {"CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"}
        if log_level not in allowed_levels:
            raise ValueError(
                f"HAPE_LOG_LEVEL must be one of {sorted(allowed_levels)}. Value is '{log_level_raw}'."
            )
        return log_level

    @staticmethod
    def get_enable_log_file() -> bool:
        raw_value = os.getenv("HAPE_ENABLE_LOG_FILE", "0")
        return ValidationUtils.validate_bool("HAPE_ENABLE_LOG_FILE", raw_value)

    @staticmethod
    def get_log_file_path() -> str:
        log_file = os.getenv("HAPE_LOG_FILE", os.path.expanduser("~/.hape/hape.log"))
        return ValidationUtils.require_string("HAPE_LOG_FILE", log_file)

    @staticmethod
    def get_exporter_host() -> str:
        host = Config._get_config_value_with_default("HAPE_EXPORTER_HOST", "0.0.0.0")
        return ValidationUtils.require_string("HAPE_EXPORTER_HOST", host)

    @staticmethod
    def get_exporter_port() -> int:
        port = Config._get_config_int_with_default("HAPE_EXPORTER_PORT", 9117)
        ValidationUtils.validate_positive_int("HAPE_EXPORTER_PORT", port)
        return port

    @staticmethod
    def get_exporter_refresh_seconds() -> int:
        refresh_seconds = Config._get_config_int_with_default("HAPE_EXPORTER_REFRESH_SECONDS", 300)
        ValidationUtils.validate_positive_int("HAPE_EXPORTER_REFRESH_SECONDS", refresh_seconds)
        return refresh_seconds

    @staticmethod
    def get_edc_kube_context() -> str:
        kube_context = Config._get_optional_config_value("HAPE_EDC_KUBE_CONTEXT")
        if kube_context is None:
            return ""
        return ValidationUtils.require_string("HAPE_EDC_KUBE_CONTEXT", kube_context)

    @staticmethod
    def get_edc_aws_profile() -> str:
        aws_profile = Config._get_optional_config_value("HAPE_EDC_AWS_PROFILE")
        if aws_profile is None:
            return ""
        return ValidationUtils.require_string("HAPE_EDC_AWS_PROFILE", aws_profile)

    @staticmethod
    def get_edc_ignored_namespaces_csv() -> str:
        default_value = "kube-system,kube-node-lease,kube-public,local-path-storage"
        ignored_namespaces_csv = Config._get_config_value_with_default("HAPE_EDC_IGNORED_NAMESPACES", default_value)
        return ValidationUtils.require_string("HAPE_EDC_IGNORED_NAMESPACES", ignored_namespaces_csv)




ERROR_MESSAGES = {
    "CSV_JSON_SOURCE_REQUIRED": "Provide one JSON source: --json-file or --json-content.",
    "CSV_JSON_SOURCE_CONFLICT": "Use either --json-file or --json-content, not both.",
    "CSV_OUTPUT_FILE_REQUIRED": "Output file path is required.",
    "CSV_INPUT_FILE_REQUIRED": "CSV input file path is required.",
    "CSV_INPUT_FILE_NOT_FOUND": "CSV file not found: {csv_path}",
    "CSV_JSON_FILE_NOT_FOUND": "JSON file not found: {json_path}",
    "CSV_DELIMITER_INVALID": "CSV delimiter must be exactly one character.",
    "CSV_INVALID_JSON": "Invalid JSON input.",
    "CSV_JSON_ARRAY_REQUIRED": "JSON input must be an array of objects.",
    "CSV_JSON_ARRAY_EMPTY": "JSON input array is empty.",
    "CSV_JSON_ROW_OBJECT_REQUIRED": "Each JSON array item must be an object.",
    "CSV_WRITE_FAILED": "Failed to write CSV output: {output_path}",
    "CSV_READ_FAILED": "Failed to read CSV input: {csv_path}",
    "CSV_WRITE_JSON_FAILED": "Failed to write JSON output: {output_path}",
}


def get_csv_error_message(message_key: str, **kwargs: str) -> str:
    template = ERROR_MESSAGES.get(message_key, "Unknown CSV error.")
    return template.format(**kwargs)


if __name__ == "__main__":
    print(get_csv_error_message("CSV_JSON_SOURCE_REQUIRED"))



from collections import Counter
from typing import Any, Dict, List, Optional

from clients.gitlab_client import GitLabClient
from core.errors.exceptions import HapeExternalError, HapeValidationError
from core.errors.messages.gitlab_error_messages import get_gitlab_error_message
from core.logging import LocalLogging
from utils.datetime_utils import DatetimeUtils


class GitLabService:
    def __init__(self, gitlab_client: Optional[GitLabClient] = None) -> None:
        self.gitlab_client = gitlab_client or GitLabClient()
        self.logger = LocalLogging.get_logger("hape.gitlab_service")

    def clone_group_projects(self, group_id: int, clone_dir: str) -> None:
        self.logger.debug(f"clone_group_projects(group_id={group_id}, clone_dir={clone_dir})")
        try:
            self.gitlab_client.clone_group_projects(group_id, clone_dir)
        except Exception as exc:
            raise HapeExternalError(
                code="GITLAB_CLONE_FAILED",
                message=get_gitlab_error_message(
                    "GITLAB_CLONE_FAILED",
                    group_id=group_id,
                ),
            ) from exc
        self.logger.info(
            f"clone_group_projects completed for group_id={group_id}"
        )

    def get_group_projects(self, group_id: int) -> List[Dict[str, Any]]:
        self.logger.debug(f"get_group_projects(group_id={group_id})")
        try:
            return self.gitlab_client.get_group_projects(group_id)
        except Exception as exc:
            raise HapeExternalError(
                code="GITLAB_GET_GROUP_PROJECTS_FAILED",
                message=get_gitlab_error_message(
                    "GITLAB_GET_GROUP_PROJECTS_FAILED",
                    group_id=group_id,
                ),
            ) from exc

    def get_group_merge_requests(self, group_id: int, created_after: str, state: str = "all", author_username: Optional[str] = None) -> List[Dict[str, Any]]:
        self.logger.debug(f"get_group_merge_requests(group_id={group_id}, created_after={created_after}, state={state}, author_username={author_username})")
        if not created_after or not created_after.strip():
            raise HapeValidationError(
                code="GITLAB_CREATED_AFTER_REQUIRED",
                message=get_gitlab_error_message("GITLAB_CREATED_AFTER_REQUIRED"),
            )
        try:
            created_after_normalized = DatetimeUtils.normalize_date_format(created_after)
        except ValueError as exc:
            raise HapeValidationError(
                code="GITLAB_INVALID_CREATED_AFTER",
                message=get_gitlab_error_message(
                    "GITLAB_INVALID_CREATED_AFTER",
                    created_after=created_after,
                ),
            ) from exc
        created_after_isoformat = DatetimeUtils.date_to_isoformat(created_after_normalized)

        try:
            merge_requests = self.gitlab_client.get_group_merge_requests(
                group_id=group_id,
                created_after=created_after_isoformat,
                state=state,
                author_username=author_username,
            )
        except Exception as exc:
            raise HapeExternalError(
                code="GITLAB_GET_GROUP_MR_FAILED",
                message=get_gitlab_error_message(
                    "GITLAB_GET_GROUP_MR_FAILED",
                    group_id=group_id,
                ),
            ) from exc
        self.logger.debug(f"get_group_merge_requests fetched {len(merge_requests)} merge requests for group_id={group_id}")
        return merge_requests

    def count_merge_requests_per_day(self, created_after: str, group_id: Optional[int] = None, project_id: Optional[int] = None, username: Optional[str] = None) -> Counter[str]:
        self.logger.debug(f"count_merge_requests_per_day(created_after={created_after}, group_id={group_id}, project_id={project_id}, username={username})")
        if group_id is not None and project_id is not None:
            raise HapeValidationError(
                code="GITLAB_GROUP_AND_PROJECT_EXCLUSIVE",
                message=get_gitlab_error_message("GITLAB_GROUP_AND_PROJECT_EXCLUSIVE"),
            )

        if group_id is None and project_id is None:
            raise HapeValidationError(
                code="GITLAB_GROUP_OR_PROJECT_REQUIRED",
                message=get_gitlab_error_message("GITLAB_GROUP_OR_PROJECT_REQUIRED"),
            )
        if not created_after or not created_after.strip():
            raise HapeValidationError(
                code="GITLAB_CREATED_AFTER_REQUIRED",
                message=get_gitlab_error_message("GITLAB_CREATED_AFTER_REQUIRED"),
            )
        try:
            created_after_normalized = DatetimeUtils.normalize_date_format(created_after)
        except ValueError as exc:
            raise HapeValidationError(
                code="GITLAB_INVALID_CREATED_AFTER",
                message=get_gitlab_error_message(
                    "GITLAB_INVALID_CREATED_AFTER",
                    created_after=created_after,
                ),
            ) from exc

        if project_id is not None:
            created_after_isoformat = DatetimeUtils.date_to_isoformat(created_after_normalized)
            try:
                merge_requests = self.gitlab_client.get_project_merge_requests(
                    project_id=project_id,
                    created_after=created_after_isoformat,
                    state="all",
                    author_username=username,
                )
            except Exception as exc:
                raise HapeExternalError(
                    code="GITLAB_GET_PROJECT_MR_FAILED",
                    message=get_gitlab_error_message(
                        "GITLAB_GET_PROJECT_MR_FAILED",
                        project_id=project_id,
                    ),
                ) from exc
        else:
            merge_requests = self.get_group_merge_requests(
                group_id=group_id,
                created_after=created_after_normalized,
                state="all",
                author_username=username,
            )

        created_dates: List[str] = []
        for merge_request in merge_requests:
            created_at = merge_request.get("created_at")
            if not created_at:
                continue
            created_date = DatetimeUtils.parse_iso_datetime(created_at).date().isoformat()
            created_dates.append(created_date)

        counts = Counter(created_dates)
        self.logger.info(f"count_merge_requests_per_day completed with {sum(counts.values())} merge requests across {len(counts.keys())} days")
        return counts




import argparse
import json
from typing import Any

from services.markdown_service import MarkdownService


class MarkdownCommands:
    @staticmethod
    def register(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            "markdown",
            help="markdown table import/export operations.",
        )
        parser.set_defaults(func=MarkdownCommands.run_help, parser=parser)
        markdown_subparsers = parser.add_subparsers(
            dest="markdown_command",
            metavar="command",
        )
        markdown_subparsers.required = False

        export_parser = markdown_subparsers.add_parser(
            "export-tables-to-csv",
            help="export markdown pipe tables to CSV files.",
        )
        export_parser.add_argument(
            "--md-file",
            required=True,
            default=None,
            help="path to markdown file input.",
        )
        export_parser.add_argument(
            "--output-dir",
            required=True,
            default=None,
            help="directory for generated CSV files.",
        )
        export_parser.add_argument(
            "--delimiter",
            required=False,
            default=",",
            help="CSV delimiter (default: ,).",
        )
        export_parser.set_defaults(func=MarkdownCommands.run_export_tables_to_csv)

        import_parser = markdown_subparsers.add_parser(
            "import-csv-table",
            help="append CSV content as markdown table.",
        )
        import_parser.add_argument(
            "--csv-file",
            required=True,
            default=None,
            help="path to CSV file input.",
        )
        import_parser.add_argument(
            "--md-file",
            required=True,
            default=None,
            help="path to markdown file target.",
        )
        import_parser.add_argument(
            "--delimiter",
            required=False,
            default=",",
            help="CSV delimiter (default: ,).",
        )
        import_parser.add_argument(
            "--table-title",
            required=False,
            default=None,
            help="optional markdown heading before table content.",
        )
        import_parser.set_defaults(func=MarkdownCommands.run_import_csv_table)

    @staticmethod
    def run_export_tables_to_csv(args: Any) -> None:
        markdown_service = MarkdownService()
        output_files = markdown_service.export_tables_to_csv(
            markdown_file=args.md_file,
            output_dir=args.output_dir,
            delimiter=args.delimiter,
        )
        print(json.dumps(output_files, indent=4))

    @staticmethod
    def run_import_csv_table(args: Any) -> None:
        markdown_service = MarkdownService()
        output_path = markdown_service.import_csv_table(
            csv_file=args.csv_file,
            markdown_file=args.md_file,
            delimiter=args.delimiter,
            table_title=args.table_title,
        )
        print(output_path)

    @staticmethod
    def run_help(args: Any) -> None:
        args.parser.print_help()



import argparse
import json
from typing import Any

from core.logging import LocalLogging
from services.confluence_service import ConfluenceService


class ConfluenceCommands:
    @staticmethod
    def register(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            "confluence",
            help="confluence page operations.",
        )
        parser.set_defaults(func=ConfluenceCommands.run_help, parser=parser)
        confluence_subparsers = parser.add_subparsers(
            dest="confluence_command",
            metavar="command",
        )
        confluence_subparsers.required = False

        get_parser = confluence_subparsers.add_parser(
            "get-page",
            help="fetch a confluence page by ID.",
        )
        get_parser.add_argument(
            "--page-id",
            required=True,
            default=None,
            help="page ID (e.g., 123456).",
        )
        get_parser.add_argument(
            "--expand",
            required=False,
            default="body.storage",
            help="confluence expand parameter (default: body.storage).",
        )
        get_parser.set_defaults(func=ConfluenceCommands.run_get_page)

        create_parser = confluence_subparsers.add_parser(
            "create-page",
            help="create a confluence page under a parent page.",
        )
        create_parser.add_argument(
            "--parent-page-id",
            required=True,
            default=None,
            help="parent page ID (e.g., 123456).",
        )
        create_parser.add_argument(
            "--page-title",
            required=True,
            default=None,
            help="page title.",
        )
        create_parser.add_argument(
            "--page-body",
            required=True,
            default=None,
            help="page body (storage format).",
        )
        create_parser.add_argument(
            "--space-key",
            required=True,
            default=None,
            help="confluence space key.",
        )
        create_parser.add_argument(
            "--labels",
            required=False,
            nargs="*",
            default=None,
            help="confluence labels (optional).",
        )
        create_parser.set_defaults(func=ConfluenceCommands.run_create_page)

        readme_parser = confluence_subparsers.add_parser(
            "md-to-page",
            help="create a confluence page from a markdown file.",
        )
        readme_parser.add_argument(
            "--parent-page-id",
            required=False,
            default=None,
            help="parent page ID (default: CONFLUENCE_TEST_PARENT_PAGE_ID).",
        )
        readme_parser.add_argument(
            "--md-path",
            required=False,
            default="README.md",
            help="path to markdown file (default: README.md).",
        )
        readme_parser.add_argument(
            "--page-title",
            required=False,
            default=None,
            help="page title (default: derived from README).",
        )
        readme_parser.add_argument(
            "--space-key",
            required=True,
            default=None,
            help="confluence space key.",
        )
        readme_parser.add_argument(
            "--labels",
            required=False,
            nargs="*",
            default=None,
            help="confluence labels (optional).",
        )
        readme_parser.set_defaults(func=ConfluenceCommands.run_readme_to_page)

    @staticmethod
    def run_get_page(args: Any) -> None:
        confluence_service = ConfluenceService()
        page = confluence_service.get_page(args.page_id, expand=args.expand)
        print(json.dumps(page, indent=4))

    @staticmethod
    def run_create_page(args: Any) -> None:
        confluence_service = ConfluenceService()
        page = confluence_service.create_page(
            parent_page_id=args.parent_page_id,
            page_title=args.page_title,
            page_body=args.page_body,
            space_key=args.space_key,
            labels=args.labels,
        )
        print(confluence_service.get_page_link(page["id"], args.space_key))

    @staticmethod
    def run_readme_to_page(args: Any) -> None:
        confluence_service = ConfluenceService()
        page = confluence_service.create_page_from_markdown(
            parent_page_id=args.parent_page_id,
            readme_path=args.md_path,
            page_title=args.page_title,
            space_key=args.space_key,
            labels=args.labels,
        )
        print(confluence_service.get_page_link(page["id"], args.space_key))

    @staticmethod
    def run_help(args: Any) -> None:
        args.parser.print_help()



import re

from typing import List


class JiraCommentFormatter:
    _FENCED_BLOCK_PATTERN = re.compile(
        r"```(?P<lang>[^\n`]*)\r?\n(?P<code>.*?)\r?\n```",
        re.DOTALL,
    )
    _MD_TABLE_SEPARATOR_PATTERN = re.compile(r"^\s*\|?[\s:-]+\|[\s\|:-]*$")
    _HEADING_PATTERN = re.compile(r"^(?P<level>#{1,6})\s+(?P<title>.+)$")
    _INLINE_CODE_PATTERN = re.compile(r"`([^`]+)`")

    @classmethod
    def format_markdown_to_jira_wiki(cls, markdown_text: str) -> str:
        blocks, placeholders = cls._extract_fenced_blocks(markdown_text)
        lines = cls._convert_tables(markdown_text, placeholders)
        formatted_lines: List[str] = []
        for line in lines:
            formatted_lines.append(cls._format_jira_line(line))
        return cls._restore_fenced_blocks("\n".join(formatted_lines), blocks)

    @classmethod
    def format_markdown_to_jira_wiki_with_mermaid(
        cls,
        markdown_text: str,
    ) -> tuple[str, List[str]]:
        code_blocks, mermaid_blocks, normalized, placeholders = (
            cls._extract_fenced_blocks_with_mermaid(markdown_text)
        )
        lines = cls._convert_tables(normalized, placeholders)
        formatted_lines: List[str] = []
        for line in lines:
            formatted_lines.append(cls._format_jira_line(line))
        comment_body = cls._restore_fenced_blocks(
            "\n".join(formatted_lines),
            code_blocks,
        )
        return comment_body, mermaid_blocks

    @classmethod
    def _extract_fenced_blocks(cls, markdown_text: str) -> tuple[List[str], List[str]]:
        blocks: List[str] = []
        placeholders: List[str] = []

        def replace_block(match: re.Match) -> str:
            language = match.group("lang").strip() or ""
            code = match.group("code")
            blocks.append(cls._jira_code_block(code, language))
            placeholder = f"@@JIRA_CODE_BLOCK_{len(blocks) - 1}@@"
            placeholders.append(placeholder)
            return placeholder

        normalized = cls._FENCED_BLOCK_PATTERN.sub(replace_block, markdown_text)
        return blocks, normalized.splitlines()

    @classmethod
    def _extract_fenced_blocks_with_mermaid(
        cls,
        markdown_text: str,
    ) -> tuple[List[str], List[str], str, List[str]]:
        code_blocks: List[str] = []
        mermaid_blocks: List[str] = []
        placeholders: List[str] = []

        def replace_block(match: re.Match) -> str:
            language = match.group("lang").strip()
            code = match.group("code")
            if language.lower() == "mermaid":
                placeholder = f"@@JIRA_MERMAID_{len(mermaid_blocks)}@@"
                mermaid_blocks.append(code)
            else:
                placeholder = f"@@JIRA_CODE_BLOCK_{len(code_blocks)}@@"
                code_blocks.append(cls._jira_code_block(code, language or ""))
            placeholders.append(placeholder)
            return placeholder

        normalized = cls._FENCED_BLOCK_PATTERN.sub(replace_block, markdown_text)
        return code_blocks, mermaid_blocks, normalized, placeholders

    @classmethod
    def _restore_fenced_blocks(cls, text: str, blocks: List[str]) -> str:
        for index, block in enumerate(blocks):
            placeholder = f"@@JIRA_CODE_BLOCK_{index}@@"
            text = text.replace(placeholder, block)
        return text

    @classmethod
    def _jira_code_block(cls, code: str, language: str) -> str:
        language = language.strip()
        if language:
            return f"{{code:{language}}}\n{code}\n{{code}}"
        return f"{{code}}\n{code}\n{{code}}"

    @classmethod
    def _convert_tables(cls, markdown_text: str, placeholders: List[str]) -> List[str]:
        lines = markdown_text.splitlines()
        output: List[str] = []
        index = 0
        placeholder_set = set(placeholders)
        while index < len(lines):
            line = lines[index]
            if line in placeholder_set:
                output.append(line)
                index += 1
                continue

            if (
                "|" in line
                and index + 1 < len(lines)
                and cls._MD_TABLE_SEPARATOR_PATTERN.match(lines[index + 1])
            ):
                header_cells = cls._split_md_table_row(line)
                output.append(cls._format_jira_header_row(header_cells))
                index += 2
                while index < len(lines) and "|" in lines[index]:
                    body_cells = cls._split_md_table_row(lines[index])
                    output.append(cls._format_jira_body_row(body_cells))
                    index += 1
                continue

            output.append(line)
            index += 1
        return output

    @classmethod
    def _split_md_table_row(cls, line: str) -> List[str]:
        trimmed = line.strip().strip("|")
        return [cell.strip() for cell in trimmed.split("|")]

    @staticmethod
    def _format_jira_header_row(cells: List[str]) -> str:
        return "|| " + " || ".join(cells) + " ||"

    @staticmethod
    def _format_jira_body_row(cells: List[str]) -> str:
        return "| " + " | ".join(cells) + " |"

    @classmethod
    def _format_jira_line(cls, line: str) -> str:
        if line.startswith("@@JIRA_CODE_BLOCK_") or line.startswith("@@JIRA_MERMAID_"):
            return line
        heading_match = cls._HEADING_PATTERN.match(line)
        if heading_match:
            level = len(heading_match.group("level"))
            title = heading_match.group("title").strip()
            return f"h{level}. {title}"

        if line.lstrip().startswith(("- ", "* ")):
            return cls._format_jira_list_item(line, bullet=True)

        ordered_match = re.match(r"^(?P<indent>\s*)\d+\.\s+(?P<item>.+)$", line)
        if ordered_match:
            return cls._format_jira_list_item(line, bullet=False)

        return cls._INLINE_CODE_PATTERN.sub(r"{{\1}}", line)

    @staticmethod
    def _format_jira_list_item(line: str, bullet: bool) -> str:
        indent = len(line) - len(line.lstrip(" "))
        level = max(1, indent // 2 + 1)
        marker = "*" if bullet else "#"
        content = line.lstrip(" ")
        content = (
            content[2:] if content.startswith(("- ", "* ")) else content.split(". ", 1)[-1]
        )
        return f"{marker * level} {content}"
